from isaaclab.envs import DirectMARLEnvCfg as __DirectMARLEnvCfg

from srb.utils.cfg import configclass

from ..env_cfg import BaseEnvCfg


@configclass
class DirectMarlEnvCfg(BaseEnvCfg, __DirectMARLEnvCfg):
    # Disable UI window by default
    ui_window_class_type: type | None = None

    def __post_init__(self):
        BaseEnvCfg.__post_init__(self)



from srb.utils.registry import register_srb_tasks

from .task import Task, TaskCfg
from .task_visual import VisualTask, VisualTaskCfg

BASE_TASK_NAME = __name__.split(".")[-1]
register_srb_tasks(
    {
        BASE_TASK_NAME: {},
        f"{BASE_TASK_NAME}_visual": {
            "entry_point": VisualTask,
            "task_cfg": VisualTaskCfg,
        },
    },
    default_entry_point=Task,
    default_task_cfg=TaskCfg,
)



from srb.core.action import (
    ActionGroup,
    JointPositionActionCfg,
    JointPositionActionGroup,
)
from srb.core.actuator import ImplicitActuatorCfg
from srb.core.asset import ArticulationCfg, Frame, Humanoid, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import rpy_to_quat
from srb.utils.nucleus import ISAACLAB_NUCLEUS_DIR


class UnitreeH1(Humanoid):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/unitree_h1",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Unitree/H1/h1.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=4,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 1.05),
            joint_pos={
                ".*_hip_yaw": 0.0,
                ".*_hip_roll": 0.0,
                ".*_hip_pitch": -0.28,
                ".*_knee": 0.79,
                ".*_ankle": -0.52,
                "torso": 0.0,
                ".*_shoulder_pitch": 0.28,
                ".*_shoulder_roll": 0.0,
                ".*_shoulder_yaw": 0.0,
                ".*_elbow": 0.52,
            },
            joint_vel={".*": 0.0},
        ),
        actuators={
            "legs": ImplicitActuatorCfg(
                joint_names_expr=[
                    ".*_hip_yaw",
                    ".*_hip_roll",
                    ".*_hip_pitch",
                    ".*_knee",
                    "torso",
                ],
                effort_limit=300.0,
                velocity_limit=100.0,
                stiffness={
                    ".*_hip_yaw": 150.0,
                    ".*_hip_roll": 150.0,
                    ".*_hip_pitch": 200.0,
                    ".*_knee": 200.0,
                    "torso": 200.0,
                },
                damping={
                    ".*_hip_yaw": 5.0,
                    ".*_hip_roll": 5.0,
                    ".*_hip_pitch": 5.0,
                    ".*_knee": 5.0,
                    "torso": 5.0,
                },
            ),
            "feet": ImplicitActuatorCfg(
                joint_names_expr=[".*_ankle"],
                effort_limit=100.0,
                velocity_limit=100.0,
                stiffness={".*_ankle": 20.0},
                damping={".*_ankle": 4.0},
            ),
            "arms": ImplicitActuatorCfg(
                joint_names_expr=[
                    ".*_shoulder_pitch",
                    ".*_shoulder_roll",
                    ".*_shoulder_yaw",
                    ".*_elbow",
                ],
                effort_limit=300.0,
                velocity_limit=100.0,
                stiffness={
                    ".*_shoulder_pitch": 40.0,
                    ".*_shoulder_roll": 40.0,
                    ".*_shoulder_yaw": 40.0,
                    ".*_elbow": 40.0,
                },
                damping={
                    ".*_shoulder_pitch": 10.0,
                    ".*_shoulder_roll": 10.0,
                    ".*_shoulder_yaw": 10.0,
                    ".*_elbow": 10.0,
                },
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="pelvis")
    frame_front_camera: Frame = Frame(
        prim_relpath="pelvis/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = "(left|right)_ankle_link"


class UnitreeG1(Humanoid):
    ## Model
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path="{ENV_REGEX_NS}/unitree_g1",
        spawn=UsdFileCfg(
            usd_path=f"{ISAACLAB_NUCLEUS_DIR}/Robots/Unitree/G1/g1.usd",
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=8,
                solver_velocity_iteration_count=4,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.74),
            joint_pos={
                ".*_hip_pitch_joint": -0.20,
                ".*_knee_joint": 0.42,
                ".*_ankle_pitch_joint": -0.23,
                ".*_elbow_pitch_joint": 0.87,
                "left_shoulder_roll_joint": 0.16,
                "left_shoulder_pitch_joint": 0.35,
                "right_shoulder_roll_joint": -0.16,
                "right_shoulder_pitch_joint": 0.35,
                "left_one_joint": 1.0,
                "right_one_joint": -1.0,
                "left_two_joint": 0.52,
                "right_two_joint": -0.52,
            },
            joint_vel={".*": 0.0},
        ),
        actuators={
            "legs": ImplicitActuatorCfg(
                joint_names_expr=[
                    ".*_hip_yaw_joint",
                    ".*_hip_roll_joint",
                    ".*_hip_pitch_joint",
                    ".*_knee_joint",
                    "torso_joint",
                ],
                effort_limit=300.0,
                velocity_limit=100.0,
                stiffness={
                    ".*_hip_yaw_joint": 150.0,
                    ".*_hip_roll_joint": 150.0,
                    ".*_hip_pitch_joint": 200.0,
                    ".*_knee_joint": 200.0,
                    "torso_joint": 200.0,
                },
                damping={
                    ".*_hip_yaw_joint": 5.0,
                    ".*_hip_roll_joint": 5.0,
                    ".*_hip_pitch_joint": 5.0,
                    ".*_knee_joint": 5.0,
                    "torso_joint": 5.0,
                },
                armature={
                    ".*_hip_.*": 0.01,
                    ".*_knee_joint": 0.01,
                    "torso_joint": 0.01,
                },
            ),
            "feet": ImplicitActuatorCfg(
                effort_limit=20.0,
                joint_names_expr=[".*_ankle_pitch_joint", ".*_ankle_roll_joint"],
                stiffness=20.0,
                damping=2.0,
                armature=0.01,
            ),
            "arms": ImplicitActuatorCfg(
                joint_names_expr=[
                    ".*_shoulder_pitch_joint",
                    ".*_shoulder_roll_joint",
                    ".*_shoulder_yaw_joint",
                    ".*_elbow_pitch_joint",
                    ".*_elbow_roll_joint",
                    ".*_five_joint",
                    ".*_three_joint",
                    ".*_six_joint",
                    ".*_four_joint",
                    ".*_zero_joint",
                    ".*_one_joint",
                    ".*_two_joint",
                ],
                effort_limit=300.0,
                velocity_limit=100.0,
                stiffness=40.0,
                damping=10.0,
                armature={
                    ".*_shoulder_.*": 0.01,
                    ".*_elbow_.*": 0.01,
                    ".*_five_joint": 0.001,
                    ".*_three_joint": 0.001,
                    ".*_six_joint": 0.001,
                    ".*_four_joint": 0.001,
                    ".*_zero_joint": 0.001,
                    ".*_one_joint": 0.001,
                    ".*_two_joint": 0.001,
                },
            ),
        },
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="pelvis")
    frame_imu: Frame | None = Frame(prim_relpath="imu_link")
    frame_front_camera: Frame = Frame(
        prim_relpath="pelvis/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = "(left|right)_ankle_roll_link"



from dataclasses import MISSING
from typing import Sequence

import torch

from srb._typing import StepReturn
from srb.core.asset import (
    Articulation,
    AssetVariant,
    Object,
    RigidObject,
    RigidObjectCfg,
)
from srb.core.env import (
    ManipulationEnv,
    ManipulationEnvCfg,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import reset_root_state_uniform
from srb.core.sensor import ContactSensor, ContactSensorCfg
from srb.utils.cfg import configclass
from srb.utils.math import (
    combine_frame_transforms,
    matrix_from_quat,
    rotmat_to_rot6d,
    scale_transform,
    subtract_frame_transforms,
)

from .asset import select_peg_in_hole_assembly

##############
### Config ###
##############


@configclass
class SceneCfg(ManipulationSceneCfg):
    peg: RigidObjectCfg = MISSING  # type: ignore
    hole: RigidObjectCfg = MISSING  # type: ignore


@configclass
class EventCfg(ManipulationEventCfg):
    randomize_object_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("peg"),
            "pose_range": {
                "x": (-0.25 - 0.025, -0.25 + 0.0125),
                "y": (-0.05, 0.05),
                "roll": (torch.pi / 2, torch.pi / 2),
                "yaw": (
                    torch.pi / 2 - torch.pi / 16,
                    torch.pi / 2 + torch.pi / 16,
                ),
            },
            "velocity_range": {},
        },
    )


@configclass
class TaskCfg(ManipulationEnvCfg):
    ## Assets
    peg: Object | AssetVariant | None = AssetVariant.DATASET

    ## Scene
    scene: SceneCfg = SceneCfg()

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 10.0
    is_finite_horizon: bool = True

    def __post_init__(self):
        super().__post_init__()

        # Task setup: Peg-in-hole assembly
        self.peg_in_hole_assembly = select_peg_in_hole_assembly(
            self,
            init_state=RigidObjectCfg.InitialStateCfg(pos=(0.5, 0.0, 0.02)),
            peg_kwargs={
                "activate_contact_sensors": True,
            },
        )
        # Scene: Peg
        self.scene.peg = self.peg_in_hole_assembly.peg.asset_cfg
        # Scene: Hole
        self.scene.hole = self.peg_in_hole_assembly.hole.asset_cfg

        # Sensor: End-effector contacts
        if isinstance(self.scene.contacts_end_effector, ContactSensorCfg):
            self.scene.contacts_end_effector.filter_prim_paths_expr = [
                self.scene.peg.prim_path
            ]

        # Update seed & number of variants for procedural assets
        self._update_procedural_assets()


############
### Task ###
############


class Task(ManipulationEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

        ## Get scene assets
        self._obj: RigidObject = self.scene["peg"]
        self._target: RigidObject = self.scene["hole"]

        ## Initialize buffers
        self._tf_pos_obj_initial = torch.zeros(
            (self.num_envs, 3), dtype=torch.float32, device=self.device
        )
        self._offset_pos_peg_ends = torch.tensor(
            self.cfg.peg_in_hole_assembly.peg.offset_pos_ends,
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1, 1)
        self._symmetry_peg = torch.tensor(
            self.cfg.peg_in_hole_assembly.peg.symmetry,
            dtype=torch.int32,
            device=self.device,
        ).repeat(self.num_envs)
        self._offset_pos_hole_bottom = torch.tensor(
            self.cfg.peg_in_hole_assembly.hole.offset_pos_bottom,
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1)
        self._offset_pos_hole_entrance = torch.tensor(
            self.cfg.peg_in_hole_assembly.hole.offset_pos_entrance,
            dtype=torch.float32,
            device=self.device,
        ).repeat(self.num_envs, 1)

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)
        self._tf_pos_obj_initial[env_ids] = self._obj.data.root_pos_w[env_ids]

    def extract_step_return(self) -> StepReturn:
        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Joints
            joint_pos_robot=self._robot.data.joint_pos,
            joint_pos_limits_robot=(
                self._robot.data.soft_joint_pos_limits
                if torch.all(torch.isfinite(self._robot.data.soft_joint_pos_limits))
                else None
            ),
            joint_pos_end_effector=self._end_effector.data.joint_pos
            if isinstance(self._end_effector, Articulation)
            else None,
            joint_pos_limits_end_effector=(
                self._end_effector.data.soft_joint_pos_limits
                if isinstance(self._end_effector, Articulation)
                and torch.all(
                    torch.isfinite(self._end_effector.data.soft_joint_pos_limits)
                )
                else None
            ),
            joint_acc_robot=self._robot.data.joint_acc,
            joint_applied_torque_robot=self._robot.data.applied_torque,
            # Kinematics
            fk_pos_end_effector=self._tf_end_effector.data.target_pos_source[:, 0, :],
            fk_quat_end_effector=self._tf_end_effector.data.target_quat_source[:, 0, :],
            # Transforms (world frame)
            tf_pos_end_effector=self._tf_end_effector.data.target_pos_w[:, 0, :],
            tf_quat_end_effector=self._tf_end_effector.data.target_quat_w[:, 0, :],
            tf_pos_obj_initial=self._tf_pos_obj_initial,
            tf_pos_obj=self._obj.data.root_pos_w,
            tf_quat_obj=self._obj.data.root_quat_w,
            tf_pos_target=self._target.data.root_pos_w,
            tf_quat_target=self._target.data.root_quat_w,
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_forces_end_effector=self._contacts_end_effector.data.net_forces_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            contact_force_matrix_end_effector=self._contacts_end_effector.data.force_matrix_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            ## Peg-in-hole descriptors
            offset_pos_peg_ends=self._offset_pos_peg_ends,
            symmetry_peg=self._symmetry_peg,
            offset_pos_hole_bottom=self._offset_pos_hole_bottom,
            offset_pos_hole_entrance=self._offset_pos_hole_entrance,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Joints
    joint_pos_robot: torch.Tensor,
    joint_pos_limits_robot: torch.Tensor | None,
    joint_pos_end_effector: torch.Tensor | None,
    joint_pos_limits_end_effector: torch.Tensor | None,
    joint_acc_robot: torch.Tensor,
    joint_applied_torque_robot: torch.Tensor,
    # Kinematics
    fk_pos_end_effector: torch.Tensor,
    fk_quat_end_effector: torch.Tensor,
    # Transforms (world frame)
    tf_pos_end_effector: torch.Tensor,
    tf_quat_end_effector: torch.Tensor,
    tf_pos_obj_initial: torch.Tensor,
    tf_pos_obj: torch.Tensor,
    tf_quat_obj: torch.Tensor,
    tf_pos_target: torch.Tensor,
    tf_quat_target: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_forces_end_effector: torch.Tensor | None,
    contact_force_matrix_end_effector: torch.Tensor | None,
    ## Peg-in-hole descriptors
    offset_pos_peg_ends: torch.Tensor,
    symmetry_peg: torch.Tensor,
    offset_pos_hole_bottom: torch.Tensor,
    offset_pos_hole_entrance: torch.Tensor,
) -> StepReturn:
    num_envs = episode_length.size(0)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Joints
    # Robot joints
    joint_pos_robot_normalized = (
        scale_transform(
            joint_pos_robot,
            joint_pos_limits_robot[:, :, 0],
            joint_pos_limits_robot[:, :, 1],
        )
        if joint_pos_limits_robot is not None
        else joint_pos_robot
    )
    # End-effector joints
    joint_pos_end_effector_normalized = (
        scale_transform(
            joint_pos_end_effector,
            joint_pos_limits_end_effector[:, :, 0],
            joint_pos_limits_end_effector[:, :, 1],
        )
        if joint_pos_end_effector is not None
        and joint_pos_limits_end_effector is not None
        else (
            joint_pos_end_effector
            if joint_pos_end_effector is not None
            else torch.empty((num_envs, 0), dtype=dtype, device=device)
        )
    )

    ## Kinematics
    fk_rotmat_end_effector = matrix_from_quat(fk_quat_end_effector)
    fk_rot6d_end_effector = rotmat_to_rot6d(fk_rotmat_end_effector)

    ## Transforms (world frame)
    # End-effector -> Object
    tf_pos_end_effector_to_obj, tf_quat_end_effector_to_obj = subtract_frame_transforms(
        t01=tf_pos_end_effector,
        q01=tf_quat_end_effector,
        t02=tf_pos_obj,
        q02=tf_quat_obj,
    )
    tf_rotmat_end_effector_to_obj = matrix_from_quat(tf_quat_end_effector_to_obj)
    tf_rot6d_end_effector_to_obj = rotmat_to_rot6d(tf_rotmat_end_effector_to_obj)

    # Peg -> Peg ends
    _tf_pos_peg_end0, _ = combine_frame_transforms(
        t01=tf_pos_obj,
        q01=tf_quat_obj,
        t12=offset_pos_peg_ends[:, 0],
    )
    _tf_pos_peg_end1, _ = combine_frame_transforms(
        t01=tf_pos_obj,
        q01=tf_quat_obj,
        t12=offset_pos_peg_ends[:, 1],
    )
    tf_pos_peg_ends = torch.stack([_tf_pos_peg_end0, _tf_pos_peg_end1], dim=1)

    # Hole -> Hole entrance | bottom
    tf_pos_hole_entrance, _ = combine_frame_transforms(
        t01=tf_pos_target,
        q01=tf_quat_target,
        t12=offset_pos_hole_entrance,
    )
    tf_pos_hole_bottom, _ = combine_frame_transforms(
        t01=tf_pos_target,
        q01=tf_quat_target,
        t12=offset_pos_hole_bottom,
    )

    # Peg ends -> Hole entrance
    tf_pos_peg_end0_to_hole_entrance, tf_quat_peg_to_hole_entrance = (
        subtract_frame_transforms(
            t01=tf_pos_peg_ends[:, 0],
            q01=tf_quat_obj,
            t02=tf_pos_hole_entrance,
            q02=tf_quat_target,
        )
    )
    tf_pos_peg_end1_to_hole_entrance, _ = subtract_frame_transforms(
        t01=tf_pos_peg_ends[:, 1],
        q01=tf_quat_obj,
        t02=tf_pos_hole_entrance,
    )
    tf_pos_peg_ends_to_hole_entrance = torch.stack(
        [tf_pos_peg_end0_to_hole_entrance, tf_pos_peg_end1_to_hole_entrance], dim=1
    )
    tf_rotmat_peg_to_hole = matrix_from_quat(tf_quat_peg_to_hole_entrance)
    tf_rot6d_peg_to_hole = rotmat_to_rot6d(tf_rotmat_peg_to_hole)

    # Peg ends -> Hole bottom
    tf_pos_peg_end0_to_hole_bottom, _ = subtract_frame_transforms(
        t01=tf_pos_peg_ends[:, 0],
        q01=tf_quat_obj,
        t02=tf_pos_hole_bottom,
    )
    tf_pos_peg_end1_to_hole_bottom, _ = subtract_frame_transforms(
        t01=tf_pos_peg_ends[:, 1],
        q01=tf_quat_obj,
        t02=tf_pos_hole_bottom,
    )
    tf_pos_peg_ends_to_hole_bottom = torch.stack(
        [tf_pos_peg_end0_to_hole_bottom, tf_pos_peg_end1_to_hole_bottom], dim=1
    )

    ## Contacts
    contact_forces_mean_robot = contact_forces_robot.mean(dim=1)
    contact_forces_mean_end_effector = (
        contact_forces_end_effector.mean(dim=1)
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )
    contact_forces_end_effector = (
        contact_forces_end_effector
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Joint torque
    WEIGHT_JOINT_TORQUE = -0.000025
    MAX_JOINT_TORQUE_PENALTY = -4.0
    penalty_joint_torque = torch.clamp_min(
        WEIGHT_JOINT_TORQUE
        * torch.sum(torch.square(joint_applied_torque_robot), dim=1),
        min=MAX_JOINT_TORQUE_PENALTY,
    )

    # Penalty: Joint acceleration
    WEIGHT_JOINT_ACCELERATION = -0.0005
    MAX_JOINT_ACCELERATION_PENALTY = -4.0
    penalty_joint_acceleration = torch.clamp_min(
        WEIGHT_JOINT_ACCELERATION * torch.sum(torch.square(joint_acc_robot), dim=1),
        min=MAX_JOINT_ACCELERATION_PENALTY,
    )

    # Penalty: Undesired robot contacts
    WEIGHT_UNDESIRED_ROBOT_CONTACTS = -1.0
    THRESHOLD_UNDESIRED_ROBOT_CONTACTS = 10.0
    penalty_undesired_robot_contacts = WEIGHT_UNDESIRED_ROBOT_CONTACTS * (
        torch.max(torch.norm(contact_forces_robot, dim=-1), dim=1)[0]
        > THRESHOLD_UNDESIRED_ROBOT_CONTACTS
    )

    # Reward: End-effector top-down orientation
    WEIGHT_TOP_DOWN_ORIENTATION = 1.0
    TANH_STD_TOP_DOWN_ORIENTATION = 0.15
    top_down_alignment = torch.sum(
        fk_rotmat_end_effector[:, :, 2]
        * torch.tensor((0.0, 0.0, -1.0), device=device)
        .unsqueeze(0)
        .expand(num_envs, 3),
        dim=1,
    )
    reward_top_down_orientation = WEIGHT_TOP_DOWN_ORIENTATION * (
        1.0 - torch.tanh((1.0 - top_down_alignment) / TANH_STD_TOP_DOWN_ORIENTATION)
    )

    # Reward: Distance | End-effector <--> Object
    WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ = 4.0
    TANH_STD_DISTANCE_END_EFFECTOR_TO_OBJ = 0.25
    reward_distance_end_effector_to_obj = WEIGHT_DISTANCE_END_EFFECTOR_TO_OBJ * (
        1.0
        - torch.tanh(
            torch.norm(tf_pos_end_effector_to_obj, dim=-1)
            / TANH_STD_DISTANCE_END_EFFECTOR_TO_OBJ
        )
    )

    # Reward: Grasp object
    WEIGHT_GRASP = 8.0
    THRESHOLD_GRASP = 5.0
    reward_grasp = (
        WEIGHT_GRASP
        * (
            torch.mean(
                torch.max(
                    torch.norm(contact_force_matrix_end_effector, dim=-1), dim=-1
                )[0],
                dim=1,
            )
            > THRESHOLD_GRASP
        )
        if contact_force_matrix_end_effector is not None
        else torch.zeros(num_envs, dtype=dtype, device=device)
    )

    # Reward: Lift object
    WEIGHT_LIFT = 4.0
    HEIGHT_OFFSET_LIFT = 0.2
    HEIGHT_SPAN_LIFT = 0.1
    TANH_STD_HEIGHT_LIFT = 0.05
    reward_lift = WEIGHT_LIFT * (
        1.0
        - torch.tanh(
            (
                torch.abs(
                    tf_pos_obj[:, 2] - tf_pos_obj_initial[:, 2] - HEIGHT_OFFSET_LIFT
                )
                - HEIGHT_SPAN_LIFT
            ).clamp(min=0.0)
            / TANH_STD_HEIGHT_LIFT
        )
    )

    # Reward: Alignment | Peg -> Hole | Primary Z axis
    WEIGHT_ALIGN_PEG_TO_HOLE_PRIMARY = 8.0
    TANH_STD_ALIGN_PEG_TO_HOLE_PRIMARY = 0.5
    _peg_to_hole_primary_axis_similarity = torch.abs(tf_rotmat_peg_to_hole[:, 2, 2])
    reward_align_peg_to_hole_primary = WEIGHT_ALIGN_PEG_TO_HOLE_PRIMARY * (
        1.0
        - torch.tanh(
            (1.0 - _peg_to_hole_primary_axis_similarity)
            / TANH_STD_ALIGN_PEG_TO_HOLE_PRIMARY
        )
    )

    # Reward: Alignment | Peg -> Hole | Secondary XY axes (affected by primary via power)
    WEIGHT_ALIGN_PEG_TO_HOLE_SECONDARY = 4.0
    TANH_STD_ALIGN_PEG_TO_HOLE_SECONDARY = 0.2
    _peg_to_hole_yaw = torch.atan2(
        tf_rotmat_peg_to_hole[:, 0, 1], tf_rotmat_peg_to_hole[:, 0, 0]
    )
    _symmetry_step = 2 * torch.pi / symmetry_peg
    _peg_to_hole_yaw_symmetric_directional = _peg_to_hole_yaw % _symmetry_step
    # Note: Lines above might result in NaN/inf when `peg_rot_symmetry_n=0` (infinite circular symmetry)
    #       However, the following `torch.where()` will handle this case
    _peg_to_hole_yaw_symmetric_normalized = torch.where(
        symmetry_peg <= 0,
        0.0,
        torch.min(
            _peg_to_hole_yaw_symmetric_directional,
            _symmetry_step - _peg_to_hole_yaw_symmetric_directional,
        )
        / (_symmetry_step / 2.0),
    )
    reward_align_peg_to_hole_secondary = WEIGHT_ALIGN_PEG_TO_HOLE_SECONDARY * (
        1.0
        - torch.tanh(
            _peg_to_hole_yaw_symmetric_normalized.pow(
                _peg_to_hole_primary_axis_similarity
            )
            / TANH_STD_ALIGN_PEG_TO_HOLE_SECONDARY
        )
    )

    # Reward: Distance | Peg -> Hole entrance (gradual)
    WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL = 8.0
    TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL = 0.16
    reward_distance_peg_to_hole_entrance_gradual = (
        WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL
        * (
            1.0
            - torch.tanh(
                torch.min(torch.norm(tf_pos_peg_ends_to_hole_entrance, dim=-1), dim=1)[
                    0
                ]
                / TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE_GRADUAL
            )
        )
    )

    # Reward: Distance | Peg -> Hole entrance
    WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE = 32.0
    TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE = 0.04
    reward_distance_peg_to_hole_entrance = WEIGHT_DISTANCE_PEG_TO_HOLE_ENTRANCE * (
        1.0
        - torch.tanh(
            torch.min(torch.norm(tf_pos_peg_ends_to_hole_entrance, dim=-1), dim=1)[0]
            / TANH_STD_DISTANCE_PEG_TO_HOLE_ENTRANCE
        )
    )

    # Reward: Distance | Peg -> Hole bottom
    WEIGHT_DISTANCE_PEG_TO_HOLE_BOTTOM = 256.0
    TANH_STD_DISTANCE_PEG_TO_HOLE_BOTTOM = 0.002
    reward_distance_peg_to_hole_bottom = WEIGHT_DISTANCE_PEG_TO_HOLE_BOTTOM * (
        1.0
        - torch.tanh(
            torch.min(torch.norm(tf_pos_peg_ends_to_hole_bottom, dim=-1), dim=1)[0]
            / TANH_STD_DISTANCE_PEG_TO_HOLE_BOTTOM
        )
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "contact_forces_mean_robot": contact_forces_mean_robot,
                "contact_forces_mean_end_effector": contact_forces_mean_end_effector,
                "tf_pos_end_effector_to_obj": tf_pos_end_effector_to_obj,
                "tf_rot6d_end_effector_to_obj": tf_rot6d_end_effector_to_obj,
                "tf_pos_peg_end0_to_hole_entrance": tf_pos_peg_end0_to_hole_entrance,
                "tf_pos_peg_end1_to_hole_entrance": tf_pos_peg_end1_to_hole_entrance,
                "tf_pos_peg_end0_to_hole_bottom": tf_pos_peg_end0_to_hole_bottom,
                "tf_pos_peg_end1_to_hole_bottom": tf_pos_peg_end1_to_hole_bottom,
                "tf_rot6d_peg_to_hole": tf_rot6d_peg_to_hole,
            },
            "state_dyn": {
                "contact_forces_robot": contact_forces_robot,
                "contact_forces_end_effector": contact_forces_end_effector,
            },
            "proprio": {
                "fk_pos_end_effector": fk_pos_end_effector,
                "fk_rot6d_end_effector": fk_rot6d_end_effector,
            },
            "proprio_dyn": {
                "joint_pos_robot_normalized": joint_pos_robot_normalized,
                "joint_pos_end_effector_normalized": joint_pos_end_effector_normalized,
                "joint_acc_robot": joint_acc_robot,
                "joint_applied_torque_robot": joint_applied_torque_robot,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_joint_torque": penalty_joint_torque,
            "penalty_joint_acceleration": penalty_joint_acceleration,
            "penalty_undesired_robot_contacts": penalty_undesired_robot_contacts,
            "reward_top_down_orientation": reward_top_down_orientation,
            "reward_distance_end_effector_to_obj": reward_distance_end_effector_to_obj,
            "reward_grasp": reward_grasp,
            "reward_lift": reward_lift,
            "reward_align_peg_to_hole_primary": reward_align_peg_to_hole_primary,
            "reward_align_peg_to_hole_secondary": reward_align_peg_to_hole_secondary,
            "reward_distance_peg_to_hole_entrance_gradual": reward_distance_peg_to_hole_entrance_gradual,
            "reward_distance_peg_to_hole_entrance": reward_distance_peg_to_hole_entrance,
            "reward_distance_peg_to_hole_bottom": reward_distance_peg_to_hole_bottom,
        },
        termination,
        truncation,
    )



from typing import Sequence

import torch

from srb._typing import StepReturn
from srb.core.asset import Articulation
from srb.core.env import (
    ManipulationEnv,
    ManipulationEnvCfg,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from srb.core.sensor import ContactSensor
from srb.utils.cfg import configclass
from srb.utils.math import matrix_from_quat, rotmat_to_rot6d, scale_transform

##############
### Config ###
##############


@configclass
class SceneCfg(ManipulationSceneCfg):
    pass


@configclass
class EventCfg(ManipulationEventCfg):
    pass


@configclass
class TaskCfg(ManipulationEnvCfg):
    ## Scene
    scene: SceneCfg = SceneCfg()

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 10.0
    is_finite_horizon: bool = False

    def __post_init__(self):
        super().__post_init__()


############
### Task ###
############


class Task(ManipulationEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def extract_step_return(self) -> StepReturn:
        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Joints
            joint_pos_robot=self._robot.data.joint_pos,
            joint_pos_limits_robot=(
                self._robot.data.soft_joint_pos_limits
                if torch.all(torch.isfinite(self._robot.data.soft_joint_pos_limits))
                else None
            ),
            joint_pos_end_effector=self._end_effector.data.joint_pos
            if isinstance(self._end_effector, Articulation)
            else None,
            joint_pos_limits_end_effector=(
                self._end_effector.data.soft_joint_pos_limits
                if isinstance(self._end_effector, Articulation)
                and torch.all(
                    torch.isfinite(self._end_effector.data.soft_joint_pos_limits)
                )
                else None
            ),
            joint_acc_robot=self._robot.data.joint_acc,
            joint_applied_torque_robot=self._robot.data.applied_torque,
            # Kinematics
            fk_pos_end_effector=self._tf_end_effector.data.target_pos_source[:, 0, :],
            fk_quat_end_effector=self._tf_end_effector.data.target_quat_source[:, 0, :],
            # Transforms (world frame)
            tf_pos_end_effector=self._tf_end_effector.data.target_pos_w[:, 0, :],
            tf_quat_end_effector=self._tf_end_effector.data.target_quat_w[:, 0, :],
            # Contacts
            contact_forces_robot=self._contacts_robot.data.net_forces_w,  # type: ignore
            contact_forces_end_effector=self._contacts_end_effector.data.net_forces_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
            contact_force_matrix_end_effector=self._contacts_end_effector.data.force_matrix_w
            if isinstance(self._contacts_end_effector, ContactSensor)
            else None,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Joints
    joint_pos_robot: torch.Tensor,
    joint_pos_limits_robot: torch.Tensor | None,
    joint_pos_end_effector: torch.Tensor | None,
    joint_pos_limits_end_effector: torch.Tensor | None,
    joint_acc_robot: torch.Tensor,
    joint_applied_torque_robot: torch.Tensor,
    # Kinematics
    fk_pos_end_effector: torch.Tensor,
    fk_quat_end_effector: torch.Tensor,
    # Transforms (world frame)
    tf_pos_end_effector: torch.Tensor,
    tf_quat_end_effector: torch.Tensor,
    # Contacts
    contact_forces_robot: torch.Tensor,
    contact_forces_end_effector: torch.Tensor | None,
    contact_force_matrix_end_effector: torch.Tensor | None,
) -> StepReturn:
    num_envs = episode_length.size(0)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Joints
    # Robot joints
    joint_pos_robot_normalized = (
        scale_transform(
            joint_pos_robot,
            joint_pos_limits_robot[:, :, 0],
            joint_pos_limits_robot[:, :, 1],
        )
        if joint_pos_limits_robot is not None
        else joint_pos_robot
    )
    # End-effector joints
    joint_pos_end_effector_normalized = (
        scale_transform(
            joint_pos_end_effector,
            joint_pos_limits_end_effector[:, :, 0],
            joint_pos_limits_end_effector[:, :, 1],
        )
        if joint_pos_end_effector is not None
        and joint_pos_limits_end_effector is not None
        else (
            joint_pos_end_effector
            if joint_pos_end_effector is not None
            else torch.empty((num_envs, 0), dtype=dtype, device=device)
        )
    )

    ## Kinematics
    fk_rotmat_end_effector = matrix_from_quat(fk_quat_end_effector)
    fk_rot6d_end_effector = rotmat_to_rot6d(fk_rotmat_end_effector)

    ## Contacts
    contact_forces_mean_robot = contact_forces_robot.mean(dim=1)
    contact_forces_mean_end_effector = (
        contact_forces_end_effector.mean(dim=1)
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )
    contact_forces_end_effector = (
        contact_forces_end_effector
        if contact_forces_end_effector is not None
        else torch.empty((num_envs, 0), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Joint torque
    WEIGHT_JOINT_TORQUE = -0.000025
    MAX_JOINT_TORQUE_PENALTY = -4.0
    penalty_joint_torque = torch.clamp_min(
        WEIGHT_JOINT_TORQUE
        * torch.sum(torch.square(joint_applied_torque_robot), dim=1),
        min=MAX_JOINT_TORQUE_PENALTY,
    )

    # Penalty: Joint acceleration
    WEIGHT_JOINT_ACCELERATION = -0.0005
    MAX_JOINT_ACCELERATION_PENALTY = -4.0
    penalty_joint_acceleration = torch.clamp_min(
        WEIGHT_JOINT_ACCELERATION * torch.sum(torch.square(joint_acc_robot), dim=1),
        min=MAX_JOINT_ACCELERATION_PENALTY,
    )

    # Penalty: Undesired robot contacts
    WEIGHT_UNDESIRED_ROBOT_CONTACTS = -1.0
    THRESHOLD_UNDESIRED_ROBOT_CONTACTS = 10.0
    penalty_undesired_robot_contacts = WEIGHT_UNDESIRED_ROBOT_CONTACTS * (
        torch.max(torch.norm(contact_forces_robot, dim=-1), dim=1)[0]
        > THRESHOLD_UNDESIRED_ROBOT_CONTACTS
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "contact_forces_mean_robot": contact_forces_mean_robot,
                "contact_forces_mean_end_effector": contact_forces_mean_end_effector,
            },
            "state_dyn": {
                "contact_forces_robot": contact_forces_robot,
                "contact_forces_end_effector": contact_forces_end_effector,
            },
            "proprio": {
                "fk_pos_end_effector": fk_pos_end_effector,
                "fk_rot6d_end_effector": fk_rot6d_end_effector,
            },
            "proprio_dyn": {
                "joint_pos_robot_normalized": joint_pos_robot_normalized,
                "joint_pos_end_effector_normalized": joint_pos_end_effector_normalized,
            },
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_joint_torque": penalty_joint_torque,
            "penalty_joint_acceleration": penalty_joint_acceleration,
            "penalty_undesired_robot_contacts": penalty_undesired_robot_contacts,
        },
        termination,
        truncation,
    )



from dataclasses import MISSING

import torch

from srb import assets
from srb.core.asset import AerialRobot, AssetVariant
from srb.core.domain import Domain
from srb.core.env import ViewerCfg
from srb.core.env.mobile.env import (
    MobileEnv,
    MobileEnvCfg,
    MobileEventCfg,
    MobileSceneCfg,
)
from srb.core.manager import EventTermCfg, SceneEntityCfg
from srb.core.mdp import reset_root_state_uniform
from srb.utils.cfg import configclass


@configclass
class AerialSceneCfg(MobileSceneCfg):
    env_spacing: float = 64.0


@configclass
class AerialEventCfg(MobileEventCfg):
    randomize_robot_state: EventTermCfg = EventTermCfg(
        func=reset_root_state_uniform,
        mode="reset",
        params={
            "asset_cfg": SceneEntityCfg("robot"),
            "pose_range": {
                "x": (-2.5, 2.5),
                "y": (-2.5, 2.5),
                "z": (0.5, 5.0),
                "yaw": (-torch.pi, torch.pi),
            },
            "velocity_range": {},
        },
    )


@configclass
class AerialEnvCfg(MobileEnvCfg):
    ## Scenario
    domain: Domain = Domain.MARS

    ## Assets
    robot: AerialRobot | AssetVariant = assets.Ingenuity()
    _robot: AerialRobot = MISSING  # type: ignore

    ## Scene
    scene: AerialSceneCfg = AerialSceneCfg()

    ## Events
    events: AerialEventCfg = AerialEventCfg()

    ## Time
    env_rate: float = 1.0 / 25.0
    agent_rate: float = 1.0 / 25.0

    ## Viewer
    viewer: ViewerCfg = ViewerCfg(
        eye=(-10.0, 0.0, 20.0), lookat=(0.0, 0.0, 0.0), origin_type="env"
    )

    def __post_init__(self):
        super().__post_init__()


class AerialEnv(MobileEnv):
    cfg: AerialEnvCfg

    def __init__(self, cfg: AerialEnvCfg, **kwargs):
        super().__init__(cfg, **kwargs)



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.robot.mobile.ground_robot import GroundRobot
from srb.core.asset.robot.mobile_manipulation.mobile_manipulator import (
    MobileManipulatorRegistry,
)
from srb.core.asset.robot.mobile_manipulation.mobile_manipulator_type import (
    MobileManipulatorType,
)

from ._metaclass import CombinedMobileManipulator


class GroundManipulator(
    CombinedMobileManipulator,
    GroundRobot,
    mobile_manipulator_entrypoint=MobileManipulatorType.GROUND,
):
    ## Model
    mobile_base: GroundRobot | None = None

    @classmethod
    def mobile_manipulator_registry(cls) -> Sequence[Type[GroundManipulator]]:
        return MobileManipulatorRegistry.registry.get(MobileManipulatorType.GROUND, [])  # type: ignore



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.common import Frame
from srb.core.asset.object.object import Object, ObjectRegistry
from srb.core.asset.object.object_type import ObjectType


class Pedestal(Object, object_entrypoint=ObjectType.PEDESTAL):
    ## Frames
    frame_manipulator_mount: Frame

    @classmethod
    def object_registry(cls) -> Sequence[Type[Pedestal]]:
        return ObjectRegistry.registry.get(ObjectType.PEDESTAL, [])  # type: ignore



from __future__ import annotations

from typing import Sequence, Type

from srb.core.asset.common import Frame
from srb.core.asset.robot.mobile.mobile_robot import MobileRobot, MobileRobotRegistry
from srb.core.asset.robot.mobile.mobile_robot_type import MobileRobotType


class OrbitalRobot(
    MobileRobot,
    mobile_robot_entrypoint=MobileRobotType.ORBITAL,
    arbitrary_types_allowed=True,
):
    ## Frames
    frame_onboard_camera: Frame

    @classmethod
    def mobile_robot_registry(cls) -> Sequence[Type[OrbitalRobot]]:
        return MobileRobotRegistry.registry.get(MobileRobotType.ORBITAL, [])  # type: ignore


class Lander(OrbitalRobot, mobile_robot_metaclass=True):
    pass



from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Dict, Iterable, List, Sequence, Tuple, Type

from srb.core.asset import ArticulationCfg
from srb.core.asset.common import Frame
from srb.core.asset.object import Tool
from srb.core.asset.robot.manipulation.manipulator_type import ManipulatorType
from srb.core.asset.robot.robot import Robot, RobotRegistry
from srb.core.asset.robot.robot_type import RobotType


class Manipulator(Robot, robot_entrypoint=RobotType.MANIPULATOR):
    ## Model
    asset_cfg: ArticulationCfg
    end_effector: Tool | None = None

    ## Frames
    frame_flange: Frame
    frame_base_camera: Frame
    frame_wrist_camera: Frame

    def __init_subclass__(
        cls,
        manipulator_entrypoint: ManipulatorType | None = None,
        manipulator_metaclass: bool = False,
        robot_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(
            robot_metaclass=(
                robot_metaclass
                or manipulator_entrypoint is not None
                or manipulator_metaclass
            ),
            **kwargs,
        )
        if manipulator_entrypoint is not None:
            assert isinstance(manipulator_entrypoint, ManipulatorType), (
                f"Class '{cls.__name__}' is marked as a manipulator entrypoint, but '{manipulator_entrypoint}' is not a valid {ManipulatorType}"
            )
            assert (
                manipulator_entrypoint not in ManipulatorRegistry.base_types.keys()
            ), (
                f"Class '{cls.__name__}' is marked as '{manipulator_entrypoint}' manipulator entrypoint, but it was already marked by '{ManipulatorRegistry.base_types[manipulator_entrypoint].__name__}'"
            )
            ManipulatorRegistry.base_types[manipulator_entrypoint] = cls
        elif manipulator_metaclass:
            ManipulatorRegistry.meta_types.append(cls)
        elif not robot_metaclass:
            for manipulator_type, base in ManipulatorRegistry.base_types.items():
                if issubclass(cls, base):
                    if manipulator_type not in ManipulatorRegistry.registry.keys():
                        ManipulatorRegistry.registry[manipulator_type] = []
                    else:
                        assert cls.name() not in (
                            manipulator.name()
                            for manipulator in ManipulatorRegistry.registry[
                                manipulator_type
                            ]
                        ), (
                            f"Cannot register multiple manipulators with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(manipulator for manipulator in ManipulatorRegistry.registry[manipulator_type] if cls.name() == manipulator.name()).__module__}:{cls.__name__}'"
                        )
                    ManipulatorRegistry.registry[manipulator_type].append(cls)
                    break

    @cached_property
    def manipulator_type(self) -> ManipulatorType:
        for manipulator_type, base in ManipulatorRegistry.base_types.items():
            if isinstance(self, base):
                return manipulator_type
        raise ValueError(
            f"Class '{self.__class__.__name__}' has unknown manipulator type"
        )

    @classmethod
    def manipulator_registry(cls) -> Sequence[Type[Manipulator]]:
        return list(ManipulatorRegistry.values_inner())

    @classmethod
    def robot_registry(cls) -> Sequence[Type[Manipulator]]:
        return RobotRegistry.registry.get(RobotType.MANIPULATOR, [])  # type: ignore


class ManipulatorRegistry:
    registry: ClassVar[Dict[ManipulatorType, List[Type[Manipulator]]]] = {}
    base_types: ClassVar[Dict[ManipulatorType, Type[Manipulator]]] = {}
    meta_types: ClassVar[List[Type[Manipulator]]] = []

    @classmethod
    def keys(cls) -> Iterable[ManipulatorType]:
        return cls.registry.keys()

    @classmethod
    def items(cls) -> Iterable[Tuple[ManipulatorType, Sequence[Type[Manipulator]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[Manipulator]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[Manipulator]]:
        return (
            manipulator
            for manipulators in cls.registry.values()
            for manipulator in manipulators
        )

    @classmethod
    def n_robots(cls) -> int:
        return sum(len(manipulators) for manipulators in cls.registry.values())

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            manipulator.__module__
            for manipulators in cls.registry.values()
            for manipulator in manipulators
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[Manipulator] | None:
        for manipulator in cls.values_inner():
            if manipulator.name() == name:
                return manipulator
        return None



from isaaclab.assets import *  # noqa: F403
from isaacsim.core.prims import XFormPrim  # noqa: F401

from .asset import Asset, AssetRegistry  # noqa: F401
from .asset_type import AssetType  # noqa: F401
from .asset_variant import AssetVariant  # noqa: F401
from .common import *  # noqa: F403
from .object import *  # noqa: F403
from .robot import *  # noqa: F403
from .scenery import *  # noqa: F403



from typing import Sequence

import torch

from srb._typing import StepReturn
from srb.core.action import ThrustAction
from srb.core.env import OrbitalEnv, OrbitalEnvCfg, OrbitalEventCfg, OrbitalSceneCfg
from srb.utils.cfg import configclass
from srb.utils.math import matrix_from_quat, rotmat_to_rot6d

##############
### Config ###
##############


@configclass
class SceneCfg(OrbitalSceneCfg):
    pass


@configclass
class EventCfg(OrbitalEventCfg):
    pass


@configclass
class TaskCfg(OrbitalEnvCfg):
    ## Scene
    scene: SceneCfg = SceneCfg()

    ## Events
    events: EventCfg = EventCfg()

    ## Time
    episode_length_s: float = 10.0
    is_finite_horizon: bool = False

    def __post_init__(self):
        super().__post_init__()


############
### Task ###
############


class Task(OrbitalEnv):
    cfg: TaskCfg

    def __init__(self, cfg: TaskCfg, **kwargs):
        super().__init__(cfg, **kwargs)

    def _reset_idx(self, env_ids: Sequence[int]):
        super()._reset_idx(env_ids)

    def extract_step_return(self) -> StepReturn:
        ## Get remaining fuel (if applicable)
        if self._thrust_action_term_key:
            thrust_action_term: ThrustAction = self.action_manager._terms[  # type: ignore
                self._thrust_action_term_key
            ]
            remaining_fuel = (
                thrust_action_term.remaining_fuel / thrust_action_term.cfg.fuel_capacity
            ).unsqueeze(-1)
        else:
            remaining_fuel = None

        return _compute_step_return(
            ## Time
            episode_length=self.episode_length_buf,
            max_episode_length=self.max_episode_length,
            truncate_episodes=self.cfg.truncate_episodes,
            ## Actions
            act_current=self.action_manager.action,
            act_previous=self.action_manager.prev_action,
            ## States
            # Root
            tf_pos_robot=self._robot.data.root_pos_w,
            tf_quat_robot=self._robot.data.root_quat_w,
            vel_lin_robot=self._robot.data.root_lin_vel_b,
            vel_ang_robot=self._robot.data.root_ang_vel_b,
            # IMU
            imu_lin_acc=self._imu_robot.data.lin_acc_b,
            imu_ang_vel=self._imu_robot.data.ang_vel_b,
            # Fuel
            remaining_fuel=remaining_fuel,
        )


@torch.jit.script
def _compute_step_return(
    *,
    ## Time
    episode_length: torch.Tensor,
    max_episode_length: int,
    truncate_episodes: bool,
    ## Actions
    act_current: torch.Tensor,
    act_previous: torch.Tensor,
    ## States
    # Root
    tf_pos_robot: torch.Tensor,
    tf_quat_robot: torch.Tensor,
    vel_lin_robot: torch.Tensor,
    vel_ang_robot: torch.Tensor,
    # IMU
    imu_lin_acc: torch.Tensor,
    imu_ang_vel: torch.Tensor,
    # Fuel
    remaining_fuel: torch.Tensor | None,
) -> StepReturn:
    num_envs = episode_length.size(0)
    dtype = episode_length.dtype
    device = episode_length.device

    ############
    ## States ##
    ############
    ## Root
    tf_rotmat_robot = matrix_from_quat(tf_quat_robot)
    tf_rot6d_robot = rotmat_to_rot6d(tf_rotmat_robot)

    ## Fuel
    remaining_fuel = (
        remaining_fuel
        if remaining_fuel is not None
        else torch.ones((num_envs, 1), dtype=dtype, device=device)
    )

    #############
    ## Rewards ##
    #############
    # Penalty: Action rate
    WEIGHT_ACTION_RATE = -0.05
    penalty_action_rate = WEIGHT_ACTION_RATE * torch.sum(
        torch.square(act_current - act_previous), dim=1
    )

    # Penalty: Fuel consumption
    WEIGHT_FUEL_CONSUMPTION = -2.0
    penalty_fuel_consumption = WEIGHT_FUEL_CONSUMPTION * torch.square(
        1.0 - remaining_fuel.squeeze(-1)
    )

    ##################
    ## Terminations ##
    ##################
    # No termination condition
    termination = torch.zeros(num_envs, dtype=torch.bool, device=device)
    # Truncation
    truncation = (
        episode_length >= max_episode_length
        if truncate_episodes
        else torch.zeros(num_envs, dtype=torch.bool, device=device)
    )

    return StepReturn(
        {
            "state": {
                "tf_pos_robot": tf_rot6d_robot,
                "tf_rot6d_robot": tf_rot6d_robot,
                "vel_lin_robot": vel_lin_robot,
                "vel_ang_robot": vel_ang_robot,
            },
            # "state_dyn": {},
            "proprio": {
                "imu_lin_acc": imu_lin_acc,
                "imu_ang_vel": imu_ang_vel,
                "remaining_fuel": remaining_fuel,
            },
            # "proprio_dyn": {},
        },
        {
            "penalty_action_rate": penalty_action_rate,
            "penalty_fuel_consumption": penalty_fuel_consumption,
        },
        termination,
        truncation,
    )



from typing import ClassVar, Sequence

import simforge_foundry

from srb.core.asset import AssetBaseCfg, Terrain
from srb.core.domain import Domain
from srb.core.sim import CollisionPropertiesCfg, SimforgeAssetCfg


### ANCHOR: example (docs)
class MarsSurface(Terrain):
    ## Scenario - The asset is suitable for the Mars domain
    DOMAINS: ClassVar[Sequence[Domain]] = (Domain.MARS,)

    ## Model - Static asset
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/mars_surface",
        ## Spawner procedurally generates SimForge models
        spawn=SimforgeAssetCfg(
            assets=[simforge_foundry.MarsSurface()],
            collision_props=CollisionPropertiesCfg(),
        ),
    )
    ### ANCHOR_END: example (docs)


class MoonSurface(Terrain):
    ## Scenario
    DOMAINS: ClassVar[Sequence[Domain]] = (Domain.MOON,)

    ## Model
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/moon_surface",
        spawn=SimforgeAssetCfg(
            assets=[simforge_foundry.MoonSurface()],
            collision_props=CollisionPropertiesCfg(),
        ),
    )



from __future__ import annotations

from functools import cached_property
from typing import ClassVar, Dict, Iterable, List, Sequence, Tuple, Type

from srb.core.asset.common import Frame
from srb.core.asset.object.payload import Payload
from srb.core.asset.robot.mobile.mobile_robot_type import MobileRobotType
from srb.core.asset.robot.robot import Robot, RobotRegistry
from srb.core.asset.robot.robot_type import RobotType


class MobileRobot(Robot, robot_entrypoint=RobotType.MOBILE_ROBOT):
    ## Model
    payload: Payload | None = None

    ## Frames
    frame_imu: Frame | None = None
    frame_payload_mount: Frame
    frame_manipulator_mount: Frame

    def __init_subclass__(
        cls,
        mobile_robot_entrypoint: MobileRobotType | None = None,
        mobile_robot_metaclass: bool = False,
        robot_metaclass: bool = False,
        **kwargs,
    ):
        super().__init_subclass__(
            robot_metaclass=(
                robot_metaclass
                or mobile_robot_entrypoint is not None
                or mobile_robot_metaclass
            ),
            **kwargs,
        )
        if mobile_robot_entrypoint is not None:
            assert isinstance(mobile_robot_entrypoint, MobileRobotType), (
                f"Class '{cls.__name__}' is marked as a mobile robot entrypoint, but '{mobile_robot_entrypoint}' is not a valid {MobileRobotType}"
            )
            assert (
                mobile_robot_entrypoint not in MobileRobotRegistry.base_types.keys()
            ), (
                f"Class '{cls.__name__}' is marked as '{mobile_robot_entrypoint}' mobile robot entrypoint, but it was already marked by '{MobileRobotRegistry.base_types[mobile_robot_entrypoint].__name__}'"
            )
            MobileRobotRegistry.base_types[mobile_robot_entrypoint] = cls
        elif mobile_robot_metaclass:
            MobileRobotRegistry.meta_types.append(cls)
        elif not robot_metaclass:
            for mobile_robot_type, base in MobileRobotRegistry.base_types.items():
                if issubclass(cls, base):
                    if mobile_robot_type not in MobileRobotRegistry.registry.keys():
                        MobileRobotRegistry.registry[mobile_robot_type] = []
                    else:
                        assert cls.name() not in (
                            mobile_robot.name()
                            for mobile_robot in MobileRobotRegistry.registry[
                                mobile_robot_type
                            ]
                        ), (
                            f"Cannot register multiple mobile robots with an identical name: '{cls.__module__}:{cls.__name__}' already exists as '{next(mobile_robot for mobile_robot in MobileRobotRegistry.registry[mobile_robot_type] if cls.name() == mobile_robot.name()).__module__}:{cls.__name__}'"
                        )
                    MobileRobotRegistry.registry[mobile_robot_type].append(cls)
                    break

    @cached_property
    def mobile_robot_type(self) -> MobileRobotType:
        for mobile_robot_type, base in MobileRobotRegistry.base_types.items():
            if isinstance(self, base):
                return mobile_robot_type
        raise ValueError(
            f"Class '{self.__class__.__name__}' has unknown mobile robot type"
        )

    @classmethod
    def mobile_robot_registry(cls) -> Sequence[Type[MobileRobot]]:
        return list(MobileRobotRegistry.values_inner())

    @classmethod
    def robot_registry(cls) -> Sequence[Type[MobileRobot]]:
        return RobotRegistry.registry.get(RobotType.MOBILE_ROBOT, [])  # type: ignore


class MobileRobotRegistry:
    registry: ClassVar[Dict[MobileRobotType, List[Type[MobileRobot]]]] = {}
    base_types: ClassVar[Dict[MobileRobotType, Type[MobileRobot]]] = {}
    meta_types: ClassVar[List[Type[MobileRobot]]] = []

    @classmethod
    def keys(cls) -> Iterable[MobileRobotType]:
        return cls.registry.keys()

    @classmethod
    def items(cls) -> Iterable[Tuple[MobileRobotType, Sequence[Type[MobileRobot]]]]:
        return cls.registry.items()

    @classmethod
    def values(cls) -> Iterable[Iterable[Type[MobileRobot]]]:
        return cls.registry.values()

    @classmethod
    def values_inner(cls) -> Iterable[Type[MobileRobot]]:
        return (
            mobile_robot
            for mobile_robots in cls.registry.values()
            for mobile_robot in mobile_robots
        )

    @classmethod
    def n_robots(cls) -> int:
        return sum(len(mobile_robots) for mobile_robots in cls.registry.values())

    @classmethod
    def registered_modules(cls) -> Iterable[str]:
        return {
            mobile_robot.__module__
            for mobile_robots in cls.registry.values()
            for mobile_robot in mobile_robots
        }

    @classmethod
    def registered_packages(cls) -> Iterable[str]:
        return {module.split(".", maxsplit=1)[0] for module in cls.registered_modules()}

    @classmethod
    def get_by_name(cls, name: str) -> Type[MobileRobot] | None:
        for mobile_robot in cls.values_inner():
            if mobile_robot.name() == name:
                return mobile_robot
        return None



from srb.core.asset import AssetBaseCfg, Frame, Pedestal, Transform
from srb.core.sim import CollisionPropertiesCfg, UsdFileCfg
from srb.utils.path import SRB_ASSETS_DIR_SRB_OBJECT


class IndustrialPedestal25(Pedestal):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/pedestal",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_OBJECT.joinpath(
                "industrial_pedestal_25cm.usdz"
            ).as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
    )

    frame_manipulator_mount: Frame = Frame(
        prim_relpath="pedestal", offset=Transform(pos=(0.0, 0.0, 0.25))
    )


class IndustrialPedestal50(Pedestal):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/pedestal",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_OBJECT.joinpath(
                "industrial_pedestal_50cm.usdz"
            ).as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
    )

    frame_manipulator_mount: Frame = Frame(
        prim_relpath="pedestal", offset=Transform(pos=(0.0, 0.0, 0.5))
    )


class IndustrialPedestal100(Pedestal):
    asset_cfg: AssetBaseCfg = AssetBaseCfg(
        prim_path="{ENV_REGEX_NS}/pedestal",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_OBJECT.joinpath(
                "industrial_pedestal_100cm.usdz"
            ).as_posix(),
            collision_props=CollisionPropertiesCfg(),
        ),
    )

    frame_manipulator_mount: Frame = Frame(
        prim_relpath="pedestal", offset=Transform(pos=(0.0, 0.0, 1.0))
    )



from isaaclab.envs import ViewerCfg  # noqa: F401
from isaaclab.scene import InteractiveScene, InteractiveSceneCfg  # noqa: F401

from .common import (  # noqa: F401
    BaseEnvCfg,
    BaseEventCfg,
    BaseSceneCfg,
    DirectEnv,
    DirectEnvCfg,
    DirectMarlEnv,
    DirectMarlEnvCfg,
    ManagedEnv,
    ManagedEnvCfg,
    VisualExt,
    VisualExtCfg,
)

# isort: split

from .manipulation import (  # noqa: F401
    ManipulationEnv,
    ManipulationEnvCfg,
    ManipulationEnvVisualExtCfg,
    ManipulationEventCfg,
    ManipulationSceneCfg,
)
from .mobile import (  # noqa: F401
    AerialEnv,
    AerialEnvCfg,
    AerialEnvVisualExtCfg,
    AerialEventCfg,
    AerialSceneCfg,
    GroundEnv,
    GroundEnvCfg,
    GroundEnvVisualExtCfg,
    GroundEventCfg,
    GroundSceneCfg,
    OrbitalEnv,
    OrbitalEnvCfg,
    OrbitalEnvVisualExtCfg,
    OrbitalEventCfg,
    OrbitalSceneCfg,
)
from .mobile_manipulation import (  # noqa: F401
    AerialManipulationEnv,
    AerialManipulationEnvCfg,
    AerialManipulationEnvVisualExtCfg,
    AerialManipulationEventCfg,
    AerialManipulationSceneCfg,
    GroundManipulationEnv,
    GroundManipulationEnvCfg,
    GroundManipulationEnvVisualExtCfg,
    GroundManipulationEventCfg,
    GroundManipulationSceneCfg,
    OrbitalManipulationEnv,
    OrbitalManipulationEnvCfg,
    OrbitalManipulationEnvVisualExtCfg,
    OrbitalManipulationEventCfg,
    OrbitalManipulationSceneCfg,
)



from pathlib import Path

import carb.settings

from srb.utils import logging

NUCLEUS_ASSET_ROOT_DIR = carb.settings.get_settings().get(
    "/persistent/isaac/asset_root/cloud"
)
NVIDIA_NUCLEUS_DIR = f"{NUCLEUS_ASSET_ROOT_DIR}/NVIDIA"
ISAAC_NUCLEUS_DIR = f"{NUCLEUS_ASSET_ROOT_DIR}/Isaac"
ISAACLAB_NUCLEUS_DIR = f"{ISAAC_NUCLEUS_DIR}/IsaacLab"


def get_local_or_nucleus_path(local_path: Path, nucleus_path: str) -> str:
    if local_path.exists():
        return local_path.as_posix()
    logging.debug(
        f"Falling back to nucleus path {nucleus_path} because {local_path} does not exist"
    )
    return nucleus_path



from .canadarm3 import Canadarm3  # noqa: F401
from .franka import Franka  # noqa: F401
from .kinova import KinovaGen3n7, KinovaJ2n6s, KinovaJ2n7s  # noqa: F401
from .so_arm100 import SOArm100D5, SOArm100D7  # noqa: F401
from .unitree import UnitreeZ1  # noqa: F401
from .universal_robots import (  # noqa: F401
    UR3,
    UR5,
    UR10,
    UR20,
    UR30,
    RandomURManipulator,
    UR3e,
    UR5e,
    UR10e,
    UR16e,
)



import datetime
import enum
import functools
import importlib
import inspect
import os
import string
import sys
from collections.abc import Callable
from dataclasses import is_dataclass
from importlib.util import find_spec
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Mapping,
    Sequence,
    Set,
    Tuple,
    Type,
    get_type_hints,
)

import gymnasium
import hydra
import yaml
from hydra.core.config_store import ConfigStore
from isaaclab.utils import configclass  # noqa: F401
from omegaconf import DictConfig, OmegaConf
from pydantic import BaseModel
from simforge import Asset as SimForgeAsset
from simforge import AssetRegistry as SimForgeAssetRegistry

from srb.core.action import ActionGroup, ActionGroupRegistry
from srb.core.asset import (
    Asset,
    AssetRegistry,
    AssetVariant,
    Manipulator,
    ManipulatorRegistry,
    MobileManipulator,
    MobileManipulatorRegistry,
    MobileRobot,
    MobileRobotRegistry,
    Object,
    ObjectRegistry,
    Payload,
    Robot,
    RobotRegistry,
    Scenery,
    SceneryRegistry,
    Tool,
)
from srb.utils import logging
from srb.utils.dict import replace_slices_with_strings, replace_strings_with_slices
from srb.utils.path import SRB_LOGS_DIR
from srb.utils.spaces import (
    replace_env_cfg_spaces_with_strings,
    replace_strings_with_env_cfg_spaces,
)

if TYPE_CHECKING:
    from srb._typing import AnyEnvCfg

SUPPORTED_FRAMEWORKS = {
    "dreamer": {"multi_algo": False},
    "sb3": {"multi_algo": True},
    "sbx": {"multi_algo": True},
    "skrl": {"multi_algo": True},
    "robomimic": {"multi_algo": True},
}
SUPPORTED_CFG_FILE_EXTENSIONS = (
    "json",
    "toml",
    "yaml",
    "yml",
)
FRAMEWORK_CFG_ENTRYPOINT_KEY = "{FRAMEWORK}_cfg"
FRAMEWORK_MULTI_ALGO_CFG_ENTRYPOINT_KEY = "{FRAMEWORK}_{ALGO}_cfg"


def parse_algo_configs(cfg_dir: str) -> Mapping[str, str]:
    algo_config = {}

    for root, _, files in os.walk(cfg_dir):
        for file in files:
            if not file.endswith(SUPPORTED_CFG_FILE_EXTENSIONS):
                continue
            file = os.path.join(root, file)

            key = _identify_config(root, file)
            if key is not None:
                algo_config[key] = file

    return algo_config


def _identify_config(root: str, file) -> str | None:
    basename = os.path.basename(file).split(".")[0]

    for framework, properties in SUPPORTED_FRAMEWORKS.items():
        algo = basename.replace(f"{framework}_", "")
        if root.endswith(framework):
            assert properties["multi_algo"]
            return FRAMEWORK_MULTI_ALGO_CFG_ENTRYPOINT_KEY.format(
                FRAMEWORK=framework, ALGO=algo
            )
        elif basename.startswith(f"{framework}"):
            if properties["multi_algo"]:
                return FRAMEWORK_MULTI_ALGO_CFG_ENTRYPOINT_KEY.format(
                    FRAMEWORK=framework, ALGO=algo
                )
            else:
                return FRAMEWORK_CFG_ENTRYPOINT_KEY.format(FRAMEWORK=framework)

    return None


def load_cfg_from_registry(
    task_name: str, entry_point_key: str, unpack_callable: bool = True
) -> "AnyEnvCfg" | Dict[str, Any]:
    # Obtain the configuration entry point
    cfg_entry_point = gymnasium.spec(task_name).kwargs.get(entry_point_key)
    # Check if entry point exists
    if cfg_entry_point is None:
        raise ValueError(
            f"Could not find configuration for the environment: '{task_name}'."
            f" Please check that the gym registry has the entry point: '{entry_point_key}'."
            f" Found: {gymnasium.spec(task_name).kwargs}."
        )
    # Parse the default config file
    if isinstance(cfg_entry_point, str) and cfg_entry_point.endswith(".yaml"):
        if os.path.exists(cfg_entry_point):
            # Absolute path for the config file
            config_file = cfg_entry_point
        else:
            # Resolve path to the module location
            mod_name, file_name = cfg_entry_point.split(":")
            mod_path = os.path.dirname(importlib.import_module(mod_name).__file__)
            # Obtain the configuration file path
            config_file = os.path.join(mod_path, file_name)
        # Load the configuration
        logging.info(f"Parsing configuration from: {config_file}")
        with open(config_file, encoding="utf-8") as f:
            cfg = yaml.full_load(f)
    else:
        if unpack_callable and callable(cfg_entry_point):
            # Resolve path to the module location
            mod_path = inspect.getfile(cfg_entry_point)
            # Load the configuration
            cfg_cls = cfg_entry_point()
        elif isinstance(cfg_entry_point, str):
            # Resolve path to the module location
            mod_name, attr_name = cfg_entry_point.split(":")
            mod = importlib.import_module(mod_name)
            cfg_cls = getattr(mod, attr_name)
        else:
            cfg_cls = cfg_entry_point
        # Load the configuration
        logging.info(f"Parsing configuration from: {cfg_entry_point}")
        cfg = cfg_cls() if unpack_callable and callable(cfg_cls) else cfg_cls
    return cfg


def stamp_dir(directory: Path, timestamp_format: str = "%Y%m%d_%H%M%S") -> Path:
    return directory.joinpath(datetime.datetime.now().strftime(timestamp_format))


def new_logdir(
    env_id: str,
    workflow: str,
    root: Path = SRB_LOGS_DIR,
    timestamp_format: str = "%Y%m%d_%H%M%S",
) -> Path:
    return stamp_dir(
        root.joinpath(env_id.removeprefix("srb/")).joinpath(workflow),
        timestamp_format=timestamp_format,
    )


def last_logdir(
    env_id: str,
    workflow: str,
    root: Path = SRB_LOGS_DIR,
    modification_time: bool = False,
) -> Path:
    logdir_parent = root.joinpath(env_id.removeprefix("srb/")).joinpath(workflow)
    if not logdir_parent.is_dir():
        raise ValueError(
            f"Path {logdir_parent} is expected to be a directory with logdirs but it "
            + ("is a file" if logdir_parent.is_file() else "does not exist")
        )

    if last_logdir := last_dir(
        directory=logdir_parent, modification_time=modification_time
    ):
        logging.debug(
            f"Selecting {last_logdir} as the last logdir"
            + (" (based on modification time)" if modification_time else "")
        )
        return last_logdir
    else:
        raise FileNotFoundError(f"Path {logdir_parent} does not contain any logdirs")


def last_dir(directory: Path, modification_time: bool = False) -> Path | None:
    assert directory.is_dir()
    if dirs := sorted(
        filter(
            lambda p: p.is_dir(),
            (directory.joinpath(child) for child in os.listdir(directory)),
        ),
        key=os.path.getmtime if modification_time else None,
        reverse=True,
    ):
        return dirs[0]
    else:
        return None


def last_file(directory: Path, modification_time: bool = False) -> Path | None:
    assert directory.is_dir()
    if files := sorted(
        filter(
            lambda p: p.is_file(),
            (directory.joinpath(child) for child in os.listdir(directory)),
        ),
        key=os.path.getmtime if modification_time else None,
        reverse=True,
    ):
        return files[0]
    else:
        return None


# TODO[mid]: Improve hydra config handling by leveraging type hints
def register_task_to_hydra(
    task_name: str, agent_cfg_entry_point: str | None = None
) -> Tuple["AnyEnvCfg", Dict[str, Any]]:
    # load the configurations
    env_cfg = load_cfg_from_registry(task_name, "task_cfg")
    # replace gymnasium spaces with strings because OmegaConf does not support them.
    # this must be done before converting the env configs to dictionary to avoid internal reinterpretations
    env_cfg = replace_env_cfg_spaces_with_strings(env_cfg)
    # convert the configs to dictionary
    env_cfg_dict = env_cfg.to_dict()

    if agent_cfg_entry_point is None:
        agent_cfg = {}
        agent_cfg_dict = {}
    else:
        agent_cfg = load_cfg_from_registry(task_name, agent_cfg_entry_point)
        if isinstance(agent_cfg, dict):
            agent_cfg_dict = agent_cfg
        else:
            agent_cfg_dict = agent_cfg.to_dict()
    cfg_dict = {"env": env_cfg_dict, "agent": agent_cfg_dict}
    # replace slices with strings because OmegaConf does not support slices
    cfg_dict = replace_slices_with_strings(cfg_dict)
    # store the configuration to Hydra
    ConfigStore.instance().store(name=task_name.rsplit("/", 1)[1], node=cfg_dict)
    return env_cfg, agent_cfg


def hydra_task_config(
    task_name: str, agent_cfg_entry_point: str | None = None
) -> Callable:
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # register the task to Hydra
            env_cfg, agent_cfg = register_task_to_hydra(
                task_name, agent_cfg_entry_point
            )

            # define the new Hydra main function
            @hydra.main(
                config_path=None,
                config_name=task_name.rsplit("/", 1)[1],
                version_base="1.3",
            )
            def hydra_main(
                hydra_env_cfg: DictConfig, env_cfg=env_cfg, agent_cfg=agent_cfg
            ):
                # convert to a native dictionary
                hydra_env_cfg = OmegaConf.to_container(
                    hydra_env_cfg,
                    # structured_config_mode=SCMode.INSTANTIATE,
                    resolve=True,
                )
                # replace string with slices because OmegaConf does not support slices
                hydra_env_cfg = replace_strings_with_slices(hydra_env_cfg)
                # update the configs with the Hydra command line arguments
                # env_cfg.from_dict(hydra_env_cfg["env"])
                env_cfg = reconstruct_object(env_cfg, hydra_env_cfg["env"])
                # replace strings that represent gymnasium spaces because OmegaConf does not support them.
                # this must be done after converting the env configs from dictionary to avoid internal reinterpretations
                replace_strings_with_env_cfg_spaces(env_cfg)
                # get agent configs
                if isinstance(agent_cfg, dict):
                    agent_cfg = hydra_env_cfg["agent"]
                else:
                    # agent_cfg.from_dict(hydra_env_cfg["agent"])
                    agent_cfg = reconstruct_object(agent_cfg, hydra_env_cfg["agent"])
                # call the original function
                func(env_cfg, agent_cfg, *args, **kwargs)

            # call the new Hydra main function
            hydra_main()

        return wrapper

    return decorator


def reconstruct_object(obj: Any, updates: Any) -> Any:
    try:
        ## String-based updates that indirectly represent a type/instance
        if (
            (not isinstance(obj, str) or isinstance(obj, enum.Enum))
            and isinstance(updates, str)
            and all(c not in string.whitespace for c in updates)
        ):
            if ":" in updates and not callable(obj):
                ## Object updated via its full module path and name
                mod_name, attr_name = updates.split(":")
                if find_spec(mod_name) is None:
                    raise ModuleNotFoundError(f"Module '{mod_name}' not found.")
                mod = importlib.import_module(mod_name)
                if not hasattr(mod, attr_name):
                    raise AttributeError(
                        f"Attribute '{attr_name}' not found in '{mod_name}'."
                    )
                attr = getattr(mod, attr_name)
                if isinstance(attr, (Type, Callable)):
                    return attr()
                else:
                    return attr
            else:
                ## Asset variant updated via variant or asset name
                if isinstance(obj, AssetVariant):
                    if variant := AssetVariant.from_str(updates):
                        # Asset variant updated via its name
                        return variant
                    elif asset_class := AssetRegistry.get_by_name(updates):
                        # Asset variant updated via asset name
                        return asset_class()  # type: ignore

                ## Registered class updated via its name
                if isinstance(obj, Asset):
                    # Asset variant
                    if variant := AssetVariant.from_str(updates):
                        return variant

                    # Scenery
                    if isinstance(obj, Scenery):
                        if scenery_class := SceneryRegistry.get_by_name(updates):
                            return scenery_class()  # type: ignore
                        else:
                            logging.warning(
                                f'Asset "{updates}" is supposed to update an instance of "{Scenery.__name__}" but it is not registered under this type'
                            )

                    # Object
                    if isinstance(obj, Object):
                        if object_class := ObjectRegistry.get_by_name(updates):
                            return object_class()  # type: ignore
                        else:
                            logging.warning(
                                f'Asset "{updates}" is supposed to update an instance of "{Object.__name__}" but it is not registered under this type'
                            )

                    # Robot
                    if isinstance(obj, Robot):
                        # Mobile manipulator
                        if isinstance(obj, MobileManipulator):
                            if (
                                mobile_manipulator_class
                                := MobileManipulatorRegistry.get_by_name(updates)
                            ):
                                return mobile_manipulator_class()  # type: ignore
                            else:
                                logging.warning(
                                    f'Asset "{updates}" is supposed to update an instance of "{MobileManipulator.__name__}" but it is not registered under this type'
                                )

                        # Manipulator
                        if isinstance(obj, Manipulator):
                            if "+" in updates:
                                manipulator_name, end_effector_name = updates.split(
                                    "+", 1
                                )

                                # Find end_effector class if specified
                                if end_effector_name:
                                    if end_effector_class := next(
                                        (
                                            end_effector_class
                                            for end_effector_class in Tool.object_registry()
                                            if end_effector_class.name()
                                            == end_effector_name
                                        ),
                                        None,
                                    ):
                                        end_effector = end_effector_class()  # type: ignore
                                    else:
                                        raise ValueError(
                                            f'Asset "{end_effector_name}" is supposed to update an instance of "{Tool.__name__}" but it is not registered under this type'
                                        )
                                else:
                                    end_effector = None

                                # Handle end_effector-only update ("+end_effector_name")
                                if not manipulator_name and end_effector:
                                    obj.end_effector = end_effector
                                    return obj

                                # Handle robot update with optional end_effector ("robot_name+end_effector_name")
                                if manipulator_name:
                                    if manipulator_class := (
                                        ManipulatorRegistry.get_by_name(
                                            manipulator_name
                                        )
                                    ):
                                        manipulator = manipulator_class()  # type: ignore
                                    else:
                                        raise ValueError(
                                            f'Asset "{manipulator_name}" is supposed to update an instance of "{Manipulator.__name__}" but it is not registered under this type'
                                        )

                                    if end_effector:
                                        manipulator.end_effector = end_effector
                                    elif manipulator.end_effector is None:
                                        manipulator.end_effector = obj.end_effector

                                    return manipulator

                            # Case 2: Format is just "robot_name"
                            if manipulator_class := ManipulatorRegistry.get_by_name(
                                updates
                            ):
                                return manipulator_class()  # type: ignore
                            else:
                                logging.warning(
                                    f'Asset "{updates}" is supposed to update an instance of "{Manipulator.__name__}" but it is not registered under this type'
                                )

                        # Mobile robot
                        if isinstance(obj, MobileRobot):
                            if "+" in updates:
                                mobile_robot_name, payload_name = updates.split("+", 1)

                                # Find payload class if specified
                                if payload_name:
                                    if payload_class := next(
                                        (
                                            payload_class
                                            for payload_class in Payload.object_registry()
                                            if payload_class.name() == payload_name
                                        ),
                                        None,
                                    ):
                                        payload = payload_class()  # type: ignore
                                    else:
                                        raise ValueError(
                                            f'Asset "{payload_name}" is supposed to update an instance of "{Payload.__name__}" but it is not registered under this type'
                                        )
                                else:
                                    payload = None

                                # Handle payload-only update ("+payload_name")
                                if not mobile_robot_name and payload:
                                    obj.payload = payload
                                    return obj

                                # Handle robot update with optional payload ("robot_name+payload_name")
                                if mobile_robot_name:
                                    if mobile_robot_class := (
                                        MobileRobotRegistry.get_by_name(
                                            mobile_robot_name
                                        )
                                    ):
                                        mobile_robot = mobile_robot_class()  # type: ignore
                                    else:
                                        raise ValueError(
                                            f'Asset "{mobile_robot_name}" is supposed to update an instance of "{MobileRobot.__name__}" but it is not registered under this type'
                                        )

                                    if payload:
                                        mobile_robot.payload = payload
                                    elif mobile_robot.payload is None:
                                        mobile_robot.payload = obj.payload

                                    return mobile_robot

                            # Case 2: Format is just "robot_name"
                            if mobile_robot_class := MobileRobotRegistry.get_by_name(
                                updates
                            ):
                                return mobile_robot_class()  # type: ignore
                            else:
                                logging.warning(
                                    f'Asset "{updates}" is supposed to update an instance of "{MobileRobot.__name__}" but it is not registered under this type'
                                )

                        # Other robot
                        if robot_class := RobotRegistry.get_by_name(updates):
                            return robot_class()  # type: ignore
                        else:
                            logging.warning(
                                f'Asset "{updates}" is supposed to update an instance of "{Robot.__name__}" but it is not registered under this type'
                            )

                    # Other asset
                    if asset_class := AssetRegistry.get_by_name(updates):
                        return asset_class()  # type: ignore
                    else:
                        raise ValueError(f'Asset "{updates}" is not registered')

                # Action group
                if isinstance(obj, ActionGroup):
                    if action_group_class := ActionGroupRegistry.get_by_name(updates):
                        return action_group_class()
                    else:
                        raise ValueError(f'Action group "{updates}" is not registered')

                # SimForge asset
                if isinstance(obj, SimForgeAsset):
                    if asset_class := SimForgeAssetRegistry.get_by_name(updates):
                        return asset_class()
                    else:
                        raise ValueError(f'Asset "{updates}" is not registered')

        # Pydantic
        if isinstance(obj, BaseModel):
            try:
                type_hints = get_type_hints(obj.__class__)
            except Exception:
                type_hints = {k: type(v) for k, v in obj.__dict__.items()}
            new_kwargs = {}
            for field_name, field_type in type_hints.items():
                if field_name.startswith("__"):
                    continue
                current_value = getattr(obj, field_name, None)
                update_value = updates.get(field_name, None)
                if update_value is not None:
                    new_kwargs[field_name] = reconstruct_object(
                        current_value, update_value
                    )
                else:
                    new_kwargs[field_name] = current_value

            return obj.__class__(**new_kwargs)

        # Dataclass
        if is_dataclass(obj):
            try:
                type_hints = get_type_hints(obj.__class__)  # type: ignore
            except Exception:
                type_hints = {k: type(v) for k, v in obj.__dict__.items()}
            new_kwargs = {}
            for field_name, field_type in type_hints.items():
                if field_name.startswith("__"):
                    continue
                current_value = getattr(obj, field_name, None)
                update_value = updates.get(field_name, None)
                if update_value is not None:
                    new_kwargs[field_name] = reconstruct_object(
                        current_value, update_value
                    )
                else:
                    new_kwargs[field_name] = current_value

            return obj.__class__(**new_kwargs)  # type: ignore

        # Enum
        if isinstance(obj, enum.Enum):
            if isinstance(updates, str):
                return obj.__class__[updates.strip().upper()]
            if isinstance(updates, Mapping) and "_name_" in updates.keys():
                return obj.__class__[updates["_name_"]]
            if updates is None and hasattr(obj, "NONE"):
                # Handle enums with "NONE" value
                return obj.__class__.NONE

        # Dict
        if isinstance(obj, Dict) and isinstance(updates, Mapping):
            obj.update(
                {
                    key: reconstruct_object(obj.get(key, None), updates.get(key, None))
                    for key in set(obj) | set(updates)
                }
            )
            return obj

        # Set
        if isinstance(obj, Set) and isinstance(updates, Iterable):
            obj.update(updates)
            return obj

        # Mapping
        if isinstance(obj, Mapping) and isinstance(updates, Mapping):
            return obj.__class__(
                (  # type: ignore
                    key,
                    reconstruct_object(obj.get(key, None), updates.get(key, None)),
                )
                for key in set(obj) | set(updates)
            )

        # Sequence
        if (isinstance(obj, Sequence) and not isinstance(obj, str)) and isinstance(
            updates, Iterable
        ):
            result = obj.__class__(
                reconstruct_object(o, u)  # type: ignore
                for o, u in zip(obj, updates)
            )
            return result

        # Callable (e.g. function)
        if callable(obj):
            return obj

        # Other types
        if not isinstance(
            obj,
            (
                str,
                int,
                float,
                bool,
                slice,
                type(None),
                Path,
            ),
        ):
            logging.warning(
                f"Type '{type(obj)}' is not explicitly handled in the object reconstruction process"
            )
        return updates if updates is not None else obj
    except Exception as e:
        overrides = ", ".join(
            f'"{arg}"' if any(c in string.whitespace for c in arg) else arg
            for arg in sys.argv[1:]
        )
        logging.critical(
            f'Failed to apply the requested override of type "{type(updates)}" to object of type "{type(obj)}" with overrides: [{overrides}]'
        )
        raise e



from dataclasses import MISSING

from srb.core.action import (
    ActionGroup,
    JointPositionActionCfg,
    JointPositionActionGroup,
)
from srb.core.actuator import ActuatorNetLSTMCfg
from srb.core.asset import ArticulationCfg, Frame, LeggedRobot, Transform
from srb.core.sim import (
    ArticulationRootPropertiesCfg,
    CollisionPropertiesCfg,
    MultiAssetSpawnerCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import rpy_to_quat
from srb.utils.nucleus import ISAACLAB_NUCLEUS_DIR


class AnymalC(LeggedRobot):
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path=MISSING,  # type: ignore
        spawn=UsdFileCfg(
            usd_path=MISSING,  # type: ignore
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.6),
            joint_pos={
                ".*HAA": 0.0,
                ".*F_HFE": 0.4,
                ".*H_HFE": -0.4,
                ".*F_KFE": -0.8,
                ".*H_KFE": 0.8,
            },
        ),
        actuators={  # type: ignore
            # "legs": DCMotorCfg(
            #     joint_names_expr=[".*HAA", ".*HFE", ".*KFE"],
            #     saturation_effort=120.0,
            #     effort_limit=80.0,
            #     velocity_limit=7.5,
            #     stiffness={".*": 40.0},
            #     damping={".*": 5.0},
            # ),
            "legs": ActuatorNetLSTMCfg(
                joint_names_expr=[".*HAA", ".*HFE", ".*KFE"],
                network_file=f"{ISAACLAB_NUCLEUS_DIR}/ActuatorNets/ANYbotics/anydrive_3_lstm_jit.pt",
                saturation_effort=120.0,
                effort_limit=80.0,
                velocity_limit=7.5,
            ),
        },
    )
    asset_cfg.prim_path = "{ENV_REGEX_NS}/anymal_c"
    asset_cfg.spawn.usd_path = (  # type: ignore
        f"{ISAACLAB_NUCLEUS_DIR}/Robots/ANYbotics/ANYmal-C/anymal_c.usd"
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(-0.075, 0.0, 0.09),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.275, 0.0, 0.09),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="base/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*FOOT"


class AnymalD(LeggedRobot):
    asset_cfg: ArticulationCfg = ArticulationCfg(
        prim_path=MISSING,  # type: ignore
        spawn=UsdFileCfg(
            usd_path=MISSING,  # type: ignore
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(
                contact_offset=0.005, rest_offset=0.0
            ),
            rigid_props=RigidBodyPropertiesCfg(
                retain_accelerations=False,
                linear_damping=0.0,
                angular_damping=0.0,
                max_linear_velocity=1000.0,
                max_angular_velocity=1000.0,
                max_depenetration_velocity=1.0,
            ),
            articulation_props=ArticulationRootPropertiesCfg(
                enabled_self_collisions=True,
                solver_position_iteration_count=4,
                solver_velocity_iteration_count=0,
            ),
        ),
        init_state=ArticulationCfg.InitialStateCfg(
            pos=(0.0, 0.0, 0.6),
            joint_pos={
                ".*HAA": 0.0,
                ".*F_HFE": 0.4,
                ".*H_HFE": -0.4,
                ".*F_KFE": -0.8,
                ".*H_KFE": 0.8,
            },
        ),
        actuators={  # type: ignore
            # "legs": DCMotorCfg(
            #     joint_names_expr=[".*HAA", ".*HFE", ".*KFE"],
            #     saturation_effort=120.0,
            #     effort_limit=80.0,
            #     velocity_limit=7.5,
            #     stiffness={".*": 40.0},
            #     damping={".*": 5.0},
            # ),
            "legs": ActuatorNetLSTMCfg(
                joint_names_expr=[".*HAA", ".*HFE", ".*KFE"],
                network_file=f"{ISAACLAB_NUCLEUS_DIR}/ActuatorNets/ANYbotics/anydrive_3_lstm_jit.pt",
                saturation_effort=120.0,
                effort_limit=80.0,
                velocity_limit=7.5,
            ),
        },
    )
    asset_cfg.prim_path = "{ENV_REGEX_NS}/anymal_d"
    asset_cfg.spawn.usd_path = (  # type: ignore
        f"{ISAACLAB_NUCLEUS_DIR}/Robots/ANYbotics/ANYmal-D/anymal_d.usd"
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(-0.075, 0.0, 0.09),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.275, 0.0, 0.09),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="base/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*FOOT"


class RandomAnymalQuadruped(LeggedRobot):
    asset_cfg: ArticulationCfg = AnymalD().asset_cfg.copy()  # type: ignore
    asset_cfg.prim_path = "{ENV_REGEX_NS}/anymal"
    asset_cfg.spawn = MultiAssetSpawnerCfg(
        random_choice=False,
        assets_cfg=(
            AnymalC().asset_cfg.spawn,  # type: ignore
            AnymalD().asset_cfg.spawn,  # type: ignore
        ),
        activate_contact_sensors=True,
    )

    ## Actions
    actions: ActionGroup = JointPositionActionGroup(
        JointPositionActionCfg(asset_name="robot", joint_names=[".*"], scale=0.5)
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(-0.075, 0.0, 0.09),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.275, 0.0, 0.09),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_front_camera: Frame = Frame(
        prim_relpath="base/camera_front",
        offset=Transform(
            pos=(-0.7675, 0.0, 1.9793),
            rot=rpy_to_quat(0.0, 15.0, -90.0),
        ),
    )

    ## Links
    regex_feet_links: str = ".*FOOT"



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from srb._typing import AnyEnv


class InterfaceBase:
    def __init__(self, env: "AnyEnv", *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        raise NotImplementedError



from .allegro_hand import AllegroHand  # noqa: F401
from .franka_hand import FrankaHand  # noqa: F401
from .kinova_gripper import Kinova300  # noqa: F401
from .robotiq import RobotiqHandE  # noqa: F401
from .scoop import ScoopRectangular, ScoopSpherical, ScoopTriangular  # noqa: F401
from .screwdriver import (  # noqa: F401
    ElectricScrewdriverM3,
    ElectricScrewdriverM5,
    ManualScrewdriverM3,
    ManualScrewdriverM5,
)
from .shadow_hand import ShadowHand  # noqa: F401
from .so_arm100 import SOArm100Gripper  # noqa: F401



from srb.core.action import ActionGroup, ThrustActionCfg, ThrustActionGroup, ThrusterCfg
from srb.core.asset import Frame, Lander, RigidObjectCfg, Transform
from srb.core.sim import (
    CollisionPropertiesCfg,
    MassPropertiesCfg,
    MeshCollisionPropertiesCfg,
    MultiAssetSpawnerCfg,
    RigidBodyPropertiesCfg,
    UsdFileCfg,
)
from srb.utils.math import deg_to_rad, rpy_to_quat
from srb.utils.path import SRB_ASSETS_DIR_SRB_ROBOT


class ApolloLander(Lander):
    ## Model
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/lander",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("lander")
            .joinpath("apollo.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                max_depenetration_velocity=5.0,
            ),
            mass_props=MassPropertiesCfg(mass=4280.0),  # Dry mass
        ),
    )

    ## Actions
    actions: ActionGroup = ThrustActionGroup(
        ThrustActionCfg(
            asset_name="robot",
            thrusters=(
                # Propulsion
                ThrusterCfg(
                    offset=(0.0, 0.0, 0.4142),
                    power=45100.0,
                ),
                # Attitude control
                ThrusterCfg(
                    offset=(1.74533, 1.82747, 4.9586),
                    direction=(0.0, 0.0, 1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(1.76876, 2.0929, 4.64178),
                    direction=(0.0, 1.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(2.00093, 1.85826, 4.64866),
                    direction=(1.0, 0.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(1.75239, 1.83069, 4.32215),
                    direction=(0.0, 0.0, -1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(1.72701, -1.74776, 4.95902),
                    direction=(0.0, 0.0, 1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(2.02006, -1.72255, 4.65853),
                    direction=(1.0, 0.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(1.80823, -2.01288, 4.65936),
                    direction=(0.0, -1.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(1.72914, -1.74409, 4.3051),
                    direction=(0.0, 0.0, -1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-1.9284, -1.65115, 5.06328),
                    direction=(0.0, 0.0, 1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-1.9069, -1.91935, 4.5559),
                    direction=(0.0, -1.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-2.11252, -1.49807, 4.55269),
                    direction=(-1.0, 0.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-1.9284, -1.66499, 4.10111),
                    direction=(0.0, 0.0, -1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-1.93109, 1.70619, 5.05008),
                    direction=(0.0, 0.0, 1.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-2.12702, 1.52566, 4.5559),
                    direction=(-1.0, 0.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-1.91524, 1.95915, 4.60795),
                    direction=(0.0, 1.0, 0.0),
                    power=445.0,
                ),
                ThrusterCfg(
                    offset=(-1.92403, 1.70298, 4.10203),
                    direction=(0.0, 0.0, -1.0),
                    power=445.0,
                ),
            ),
            # 20 seconds of full propulsion thrust, starting at 50% capacity
            fuel_capacity=0.5 * 10920.0,
            fuel_consumption_rate=(0.5 * 10920.0 / 45100.0) / 20.0,
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_onboard_camera: Frame = Frame(
        prim_relpath="base/camera_onboard",
        offset=Transform(
            pos=(0.0, -1.6, 1.1),
            rot=rpy_to_quat(0.0, 90.0, 0.0),
        ),
    )


class PeregrineLander(Lander):
    ## Model
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/lander",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("lander")
            .joinpath("peregrine.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                max_depenetration_velocity=5.0,
            ),
            mass_props=MassPropertiesCfg(mass=825.0),
        ),
    )

    ## Actions
    actions: ActionGroup = ThrustActionGroup(
        ThrustActionCfg(
            asset_name="robot",
            thrusters=(
                # Propulsion
                ThrusterCfg(
                    offset=(0.0, 0.0, 0.4448),
                    power=667.0,
                ),
                ThrusterCfg(
                    offset=(-0.21766, -0.21766, 0.4448),
                    power=667.0,
                ),
                ThrusterCfg(
                    offset=(-0.21766, 0.21766, 0.4448),
                    power=667.0,
                ),
                ThrusterCfg(
                    offset=(0.21766, -0.21766, 0.4448),
                    power=667.0,
                ),
                ThrusterCfg(
                    offset=(0.21766, 0.21766, 0.4448),
                    power=667.0,
                ),
                # Attitude control
                ThrusterCfg(
                    offset=(0.0038, -0.5947, 0.569),
                    direction=(0.0, 0.0, -1.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(-0.09, -0.647, 0.652),
                    direction=(-0.7, -1.0, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(0.0942, -0.651, 0.6524),
                    direction=(0.7, -1.0, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(0.0038, 0.5947, 0.569),
                    direction=(0.0, 0.0, -1.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(-0.09, 0.647, 0.652),
                    direction=(-0.7, 1.0, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(0.0942, 0.651, 0.6524),
                    direction=(0.7, 1.0, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(-1.05112, -0.00187, 1.7737),
                    direction=(0.0, 0.0, 1.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(-1.1074, 0.088521, 1.69),
                    direction=(-1.0, 0.7, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(-1.1034, -0.095625, 1.69067),
                    direction=(-1.0, -0.7, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(1.05112, -0.00187, 1.7737),
                    direction=(0.0, 0.0, 1.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(1.1074, 0.088521, 1.69),
                    direction=(1.0, 0.7, 0.0),
                    power=45.0,
                ),
                ThrusterCfg(
                    offset=(1.1034, -0.095625, 1.69067),
                    direction=(1.0, -0.7, 0.0),
                    power=45.0,
                ),
            ),
            fuel_capacity=0.5 * 450.0,
            fuel_consumption_rate=(0.5 * 450.0 / (5 * 667.0)) / 20.0,
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_onboard_camera: Frame = Frame(
        prim_relpath="base/camera_onboard",
        offset=Transform(
            pos=(-0.685, 0.0, 0.55),
            rot=rpy_to_quat(0.0, 90.0, 0.0),
        ),
    )


class VikramLander(Lander):
    ## Model
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/lander",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("lander")
            .joinpath("vikram.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                max_depenetration_velocity=5.0,
            ),
            mass_props=MassPropertiesCfg(mass=626.0),
        ),
    )

    ## Actions
    actions: ActionGroup = ThrustActionGroup(
        ThrustActionCfg(
            asset_name="robot",
            thrusters=(
                # Propulsion
                ThrusterCfg(
                    offset=(0.0, 0.0, 0.532305),
                    power=800.0,
                ),
                ThrusterCfg(
                    offset=(0.012183, -0.705708, 0.519355),
                    power=800.0,
                ),
                ThrusterCfg(
                    offset=(0.729607, 0.066033, 0.519355),
                    power=800.0,
                ),
                ThrusterCfg(
                    offset=(0.012183, 0.708475, 0.519355),
                    power=800.0,
                ),
                ThrusterCfg(
                    offset=(-0.738229, 0.066033, 0.519355),
                    power=800.0,
                ),
                # Attitude control
                ThrusterCfg(
                    offset=(-0.799375, 0.428285, 0.764497),
                    direction=(-1.0, 0.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(-0.807868, -0.364819, 0.760913),
                    direction=(-1.0, 0.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(-0.431845, -0.787548, 0.764497),
                    direction=(0.0, -1.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(0.418427, -0.776689, 0.760913),
                    direction=(0.0, -1.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(0.814197, -0.373962, 0.761142),
                    direction=(1.0, 0.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(0.806071, 0.419149, 0.760913),
                    direction=(1.0, 0.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(0.421697, 0.746937, 0.764498),
                    direction=(0.0, 1.0, -0.84),
                    power=58.0,
                ),
                ThrusterCfg(
                    offset=(-0.395689, 0.735653, 0.760913),
                    direction=(0.0, 1.0, -0.84),
                    power=58.0,
                ),
            ),
            fuel_capacity=0.5 * 845.0,
            fuel_consumption_rate=(0.5 * 845.0 / (5 * 800.0)) / 20.0,
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_onboard_camera: Frame = Frame(
        prim_relpath="base/camera_onboard",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 90.0, 0.0),
        ),
    )


class ResilienceLander(Lander):
    ## Model
    asset_cfg: RigidObjectCfg = RigidObjectCfg(
        prim_path="{ENV_REGEX_NS}/lander",
        spawn=UsdFileCfg(
            usd_path=SRB_ASSETS_DIR_SRB_ROBOT.joinpath("lander")
            .joinpath("resilience.usdz")
            .as_posix(),
            activate_contact_sensors=True,
            collision_props=CollisionPropertiesCfg(),
            mesh_collision_props=MeshCollisionPropertiesCfg(
                mesh_approximation="convexDecomposition"
            ),
            rigid_props=RigidBodyPropertiesCfg(
                max_depenetration_velocity=5.0,
            ),
            mass_props=MassPropertiesCfg(mass=340.0),
        ),
    )

    ## Actions
    actions: ActionGroup = ThrustActionGroup(
        ThrustActionCfg(
            asset_name="robot",
            thrusters=(
                # Propulsion
                ThrusterCfg(
                    offset=(0.0, 0.0, 0.325),
                    power=2500.0,
                    gimbal_limits=(deg_to_rad(10.0), deg_to_rad(10.0)),
                ),
                # Attitude control
                ThrusterCfg(
                    offset=(0.3435, 0.3435, 0.45816),
                    direction=(-0.25, -0.2572, -0.9334),
                    power=50.0,
                ),
                ThrusterCfg(
                    offset=(0.469217, -0.125732, 0.45816),
                    direction=(-0.293374, 0.076, -0.953),
                    power=50.0,
                ),
                ThrusterCfg(
                    offset=(-0.1257, 0.46922, 0.45816),
                    direction=(0.0977, -0.3452, -0.9334),
                    power=50.0,
                ),
                ThrusterCfg(
                    offset=(-0.3435, -0.3435, 0.45816),
                    direction=(0.25, 0.2572, -0.9334),
                    power=50.0,
                ),
                ThrusterCfg(
                    offset=(-0.469217, 0.125732, 0.45816),
                    direction=(0.293374, -0.076, -0.953),
                    power=50.0,
                ),
                ThrusterCfg(
                    offset=(0.1257, -0.46922, 0.45816),
                    direction=(-0.0977, 0.3452, -0.9334),
                    power=50.0,
                ),
            ),
            fuel_capacity=0.5 * 660.0,
            fuel_consumption_rate=(0.5 * 660.0 / (1 * 2500.0)) / 20.0,
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_onboard_camera: Frame = Frame(
        prim_relpath="base/camera_onboard",
        offset=Transform(
            pos=(0.0, 0.3, 0.55),
            rot=rpy_to_quat(0.0, 90.0, 0.0),
        ),
    )


class RandomLander(Lander):
    asset_cfg: RigidObjectCfg = ApolloLander().asset_cfg.copy()  # type: ignore
    asset_cfg.prim_path = "{ENV_REGEX_NS}/anymal"
    asset_cfg.spawn = MultiAssetSpawnerCfg(
        random_choice=False,
        assets_cfg=(
            ApolloLander().asset_cfg.spawn,  # type: ignore
            PeregrineLander().asset_cfg.spawn,  # type: ignore
            VikramLander().asset_cfg.spawn,  # type: ignore
            ResilienceLander().asset_cfg.spawn,  # type: ignore
        ),
        activate_contact_sensors=True,
    )

    ## Actions
    actions: ActionGroup = ThrustActionGroup(
        ThrustActionCfg(
            asset_name="robot",
            thrusters=(
                ThrusterCfg(
                    offset=(0.0, 0.0, 0.45),
                    power=50000.0,
                    gimbal_limits=(deg_to_rad(30.0), deg_to_rad(30.0)),
                ),
            ),
            fuel_capacity=5000.0,
            fuel_consumption_rate=(5000.0 / 50000.0) / 20.0,
        )
    )

    ## Frames
    frame_base: Frame = Frame(prim_relpath="base")
    frame_payload_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_manipulator_mount: Frame = Frame(
        prim_relpath="base",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 0.0, 0.0),
        ),
    )
    frame_onboard_camera: Frame = Frame(
        prim_relpath="base/camera_onboard",
        offset=Transform(
            pos=(0.0, 0.0, 0.0),
            rot=rpy_to_quat(0.0, 90.0, 0.0),
        ),
    )



import functools
import re

from isaaclab.utils.string import *  # noqa: F403

_REGEX_SNAKE_CASE_PATTERN = (
    re.compile(r"(.)([A-Z][a-z]+)"),
    re.compile(r"__([A-Z])"),
    re.compile(r"([a-z0-9])([A-Z])"),
)


@functools.cache
def convert_to_snake_case(input: str) -> str:
    input = _REGEX_SNAKE_CASE_PATTERN[0].sub(r"\1_\2", input)
    input = _REGEX_SNAKE_CASE_PATTERN[1].sub(r"_\1", input)
    return _REGEX_SNAKE_CASE_PATTERN[2].sub(r"\1_\2", input).lower()


_REGEX_SANITIZE_CAM_NAME = re.compile(r"cam_|camera_|sensor_")


@functools.cache
def sanitize_cam_name(name: str) -> str:
    return _REGEX_SANITIZE_CAM_NAME.sub("", name)


_REGEX_ENV_PRIM_PATH_PATTERN = re.compile(r"({ENV_REGEX_NS}|/World/envs/env_.*)")


@functools.cache
def resolve_env_prim_path(name: str, i: int) -> str:
    return _REGEX_ENV_PRIM_PATH_PATTERN.sub(f"/World/envs/env_{i}", name)



from pathlib import Path
from typing import TYPE_CHECKING, Literal

from isaacsim.simulation_app import SimulationApp
from skrl.utils.runner.torch import Runner

from srb.integrations.skrl.wrapper import SkrlEnvWrapper
from srb.utils import logging
from srb.utils.cfg import last_file, stamp_dir

if TYPE_CHECKING:
    from srb._typing import AnyEnv, AnyEnvCfg

FRAMEWORK_NAME = "skrl"


def run(
    workflow: Literal["train", "eval"],
    env: "AnyEnv",
    sim_app: SimulationApp,
    env_id: str,
    env_cfg: "AnyEnvCfg",
    agent_cfg: dict,
    logdir: Path,
    model: Path,
    continue_training: bool | None = None,
    **kwargs,
):
    # Determine checkpoint path
    if model:
        from_checkpoint = model
    elif workflow == "eval" or continue_training:
        from_checkpoint = last_file(
            logdir.joinpath("checkpoints"), modification_time=True
        )
    else:
        from_checkpoint = ""
    if from_checkpoint:
        logging.info(f"Loading model from {from_checkpoint}")

    # Special handling for eval workflow
    if workflow == "eval":
        logdir = stamp_dir(logdir.joinpath("eval"))

    # Update agent config
    agent_cfg["seed"] = env_cfg.seed
    agent_cfg["agent"]["experiment"]["directory"] = logdir.parent
    agent_cfg["agent"]["experiment"]["experiment_name"] = logdir

    # Wrap the environment
    env = SkrlEnvWrapper(env)  # type: ignore

    # Create the runner
    runner = Runner(
        env,  # type: ignore
        agent_cfg,
    )

    # Load checkpoint if needed
    if from_checkpoint:
        runner.agent.load(
            from_checkpoint,  # type: ignore
        )

    # Run the workflow
    runner.run(mode=workflow)



from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Sequence

import gymnasium
import numpy
import torch
from stable_baselines3.common.vec_env.base_vec_env import (
    VecEnv,
    VecEnvObs,
    VecEnvStepReturn,
)

if TYPE_CHECKING:
    from srb._typing import AnyEnv


class Sb3EnvWrapper(VecEnv):
    def __init__(self, env: "AnyEnv", ignore_extras: bool = True):
        # Initialize the wrapper
        self.env = env
        # Collect common information
        self.num_envs = self.unwrapped.num_envs
        self.sim_device = self.unwrapped.device
        self.render_mode = self.unwrapped.render_mode

        if ignore_extras:
            self._default_infos = [{"episode": None} for _ in range(self.num_envs)]
            self._process_extras = self._process_extras_ignore

        # Obtain gym spaces
        # Note: stable-baselines3 does not like when we have unbounded action space so
        #   we set it to some high value here. Maybe this is not general but something to think about.
        observation_space = self.unwrapped.single_observation_space  # type: ignore
        action_space = self.unwrapped.single_action_space  # type: ignore
        if isinstance(
            action_space, gymnasium.spaces.Box
        ) and not action_space.is_bounded("both"):
            action_space = gymnasium.spaces.Box(
                low=-1.0, high=1.0, shape=action_space.shape
            )

        # Initialize vec-env
        VecEnv.__init__(self, self.num_envs, observation_space, action_space)
        # Add buffer for logging episodic information
        self._ep_rew_buf = torch.zeros(self.num_envs, device=self.sim_device)
        self._ep_len_buf = torch.zeros(self.num_envs, device=self.sim_device)

        # Use the appropriate observation processing function
        if isinstance(observation_space, gymnasium.spaces.Dict):
            self._process_obs = self._process_obs_dict  # type: ignore

    def __str__(self):
        return f"<{type(self).__name__}{self.env}>"

    def __repr__(self):
        return str(self)

    @classmethod
    def class_name(cls) -> str:
        return cls.__name__

    @property
    def unwrapped(self) -> "AnyEnv":
        return self.env.unwrapped  # type: ignore

    def get_episode_rewards(self) -> Sequence[float]:
        return self._ep_rew_buf.cpu().tolist()

    def get_episode_lengths(self) -> Sequence[int]:
        return self._ep_len_buf.cpu().tolist()

    def seed(self, seed: int | None = None) -> Sequence[int | None]:
        return [self.unwrapped.seed(seed)] * self.unwrapped.num_envs  # type: ignore

    def reset(self) -> VecEnvObs:
        obs_dict, _ = self.env.reset()
        # Reset episodic information buffers
        self._ep_rew_buf.zero_()
        self._ep_len_buf.zero_()
        # Convert data types to numpy depending on backend
        return self._process_obs(obs_dict)  # type: ignore

    def step_async(self, actions):
        # Convert input to numpy array
        if not isinstance(actions, torch.Tensor):
            actions = numpy.asarray(actions)
            actions = torch.from_numpy(actions).to(
                device=self.sim_device, dtype=torch.float32
            )
        else:
            actions = actions.to(device=self.sim_device, dtype=torch.float32)
        # Convert to tensor
        self._async_actions = actions

    def step_wait(self) -> VecEnvStepReturn:
        # Record step information
        obs_dict, rew, terminated, truncated, extras = self.env.step(
            self._async_actions  # type: ignore
        )
        # Update episode un-discounted return and length
        self._ep_rew_buf += rew
        self._ep_len_buf += 1
        # Compute reset ids
        dones = terminated | truncated
        reset_ids = (dones > 0).nonzero(as_tuple=False)  # type: ignore

        # Convert data types to numpy depending on backend
        # Note: ManagerBasedRLEnv uses torch backend (by default).
        obs = self._process_obs(obs_dict)  # type: ignore
        rew = rew.detach().cpu().numpy()  # type: ignore
        terminated = terminated.detach().cpu().numpy()  # type: ignore
        truncated = truncated.detach().cpu().numpy()  # type: ignore
        dones = dones.detach().cpu().numpy()  # type: ignore
        # Convert extra information to list of dicts
        infos = self._process_extras(
            obs,  # type: ignore
            terminated,
            truncated,
            extras,
            reset_ids,  # type: ignore
        )

        # Reset info for terminated environments
        self._ep_rew_buf[reset_ids] = 0
        self._ep_len_buf[reset_ids] = 0

        return obs, rew, dones, infos  # type: ignore

    def close(self):
        self.env.close()

    def get_attr(self, attr_name: str, indices: Sequence[int] | slice | None = None):
        # Resolve indices
        if indices is None:
            indices = slice(None)
            num_indices = self.num_envs
        else:
            num_indices = len(indices)  # type: ignore
        # Obtain attribute value
        attr_val = getattr(self.env, attr_name)
        # Return the value
        if not isinstance(attr_val, torch.Tensor):
            return [attr_val] * num_indices
        else:
            return attr_val[indices].detach().cpu().numpy()

    def set_attr(
        self, attr_name: str, value: Any, indices: Sequence[int] | None = None
    ):
        raise NotImplementedError("Setting attributes is not supported.")

    def env_method(
        self,
        method_name: str,
        *method_args,
        indices: Sequence[int] | None = None,
        **method_kwargs,
    ):
        if method_name == "render":
            # Gymnasium does not support changing render mode at runtime
            return self.env.render()
        else:
            # This isn't properly implemented but it is not necessary.
            # Mostly done for completeness.
            env_method = getattr(self.env, method_name)
            return env_method(*method_args, indices=indices, **method_kwargs)

    def env_is_wrapped(self, wrapper_class, indices=None):
        raise NotImplementedError(
            "Checking if environment is wrapped is not supported."
        )

    def get_images(self):
        raise NotImplementedError("Getting images is not supported.")

    def _process_obs(self, obs: torch.Tensor) -> numpy.ndarray:
        return obs.detach().cpu().numpy()

    def _process_obs_dict(
        self, obs: Dict[str, torch.Tensor]
    ) -> Mapping[str, numpy.ndarray]:
        for key, value in obs.items():
            obs[key] = value.detach().cpu().numpy()  # type: ignore
        return obs  # type: ignore

    def _process_extras(
        self,
        obs: numpy.ndarray,
        terminated: numpy.ndarray,
        truncated: numpy.ndarray,
        extras: Dict[str, Any],
        reset_ids: numpy.ndarray,
    ) -> Sequence[Mapping[str, Any]]:
        # Create empty list of dictionaries to fill
        infos: List[Dict[str, Any]] = [
            dict.fromkeys(extras.keys()) for _ in range(self.num_envs)
        ]
        # Fill-in information for each sub-environment
        # Note: This loop becomes slow when number of environments is large.
        for idx in range(self.num_envs):
            # Fill-in episode monitoring info
            if idx in reset_ids:
                infos[idx]["episode"] = {}
                infos[idx]["episode"]["r"] = float(self._ep_rew_buf[idx])
                infos[idx]["episode"]["l"] = float(self._ep_len_buf[idx])
            else:
                infos[idx]["episode"] = None
            # Fill-in bootstrap information
            infos[idx]["TimeLimit.truncated"] = truncated[idx] and not terminated[idx]
            # Fill-in information from extras
            for key, value in extras.items():
                # 1. remap extra episodes information safely
                # 2. for others just store their values
                if key == "log":
                    # Only log this data for episodes that are terminated
                    if infos[idx]["episode"] is not None:
                        for sub_key, sub_value in value.items():
                            infos[idx]["episode"][sub_key] = sub_value
                else:
                    infos[idx][key] = value[idx]
            # Add information about terminal observation separately
            if idx in reset_ids:
                # Extract terminal observations
                if isinstance(obs, Dict):
                    terminal_obs = dict.fromkeys(obs.keys())
                    for key, value in obs.items():
                        terminal_obs[key] = value[idx]
                else:
                    terminal_obs = obs[idx]
                # Add info to dict
                infos[idx]["terminal_observation"] = terminal_obs
            else:
                infos[idx]["terminal_observation"] = None
        # Return list of dictionaries
        return infos

    def _process_extras_ignore(
        self,
        obs: numpy.ndarray,
        terminated: numpy.ndarray,
        truncated: numpy.ndarray,
        extras: Dict[str, Any],
        reset_ids: numpy.ndarray,
    ) -> Sequence[Mapping[str, Any]]:
        infos = deepcopy(self._default_infos)

        for idx in reset_ids:
            # Fill-in episode monitoring info
            infos[idx]["episode"] = {
                "r": float(self._ep_rew_buf[idx]),
                "l": float(self._ep_len_buf[idx]),
            }

            # Fill-in bootstrap information
            infos[idx]["TimeLimit.truncated"] = truncated[idx] and not terminated[idx]

            # Add information about terminal observation separately
            if isinstance(obs, dict):
                terminal_obs = {key: value[idx] for key, value in obs.items()}
            else:
                terminal_obs = obs[idx]
            infos[idx]["terminal_observation"] = terminal_obs

        return infos



from isaaclab.envs.mdp.commands import *  # noqa: F403
from isaaclab.envs.mdp.curriculums import *  # noqa: F403
from isaaclab.envs.mdp.events import *  # noqa: F403
from isaaclab.envs.mdp.observations import *  # noqa: F403
from isaaclab.envs.mdp.recorders import *  # noqa: F403
from isaaclab.envs.mdp.rewards import *  # noqa: F403
from isaaclab.envs.mdp.terminations import *  # noqa: F403

from .events import *  # noqa: F403



from dataclasses import MISSING
from typing import Callable

from srb.core.sim import ShapeCfg, SphereCfg
from srb.utils.cfg import configclass

from . import impl


@configclass
class PinCfg(ShapeCfg):
    pin_radius: float = MISSING  # type: ignore
    pin_length: float = MISSING  # type: ignore


@configclass
class ArrowCfg(ShapeCfg):
    func: Callable = impl.spawn_arrow

    tail_radius: float = MISSING  # type: ignore
    tail_length: float = MISSING  # type: ignore
    head_radius: float = MISSING  # type: ignore
    head_length: float = MISSING  # type: ignore


@configclass
class PinnedArrowCfg(PinCfg, ArrowCfg):
    func: Callable = impl.spawn_pinned_arrow


@configclass
class PinnedSphereCfg(PinCfg, SphereCfg):
    func: Callable = impl.spawn_pinned_sphere
